import autoarray as aa
import numpy as np

from autoarray.mock.mock import MockPixelizationGrid, MockRegMapper


class TestRegularizationinstance:
    def test__regularization_matrix__compare_to_regularization_util(self):

        pixel_neighbors = np.array(
            [
                [1, 3, 7, 2],
                [4, 2, 0, -1],
                [1, 5, 3, -1],
                [4, 6, 0, -1],
                [7, 1, 5, 3],
                [4, 2, 8, -1],
                [7, 3, 0, -1],
                [4, 8, 6, -1],
                [7, 5, -1, -1],
            ]
        )

        pixel_neighbors_size = np.array([4, 3, 3, 3, 4, 3, 3, 3, 2])

        pixelization_grid = MockPixelizationGrid(
            pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size
        )

        mapper = MockRegMapper(source_pixelization_grid=pixelization_grid)

        reg = aa.reg.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_mapper(mapper=mapper)

        regularization_matrix_util = aa.util.regularization.constant_regularization_matrix_from(
            coefficient=1.0,
            pixel_neighbors=pixel_neighbors,
            pixel_neighbors_size=pixel_neighbors_size,
        )

        assert (regularization_matrix == regularization_matrix_util).all()


class TestRegularizationWeighted:
    def test__weights__compare_to_regularization_util(self):

        reg = aa.reg.AdaptiveBrightness(inner_coefficient=10.0, outer_coefficient=15.0)

        pixel_signals = np.array([0.21, 0.586, 0.45])

        mapper = MockRegMapper(pixel_signals=pixel_signals)

        weights = reg.regularization_weights_from_mapper(mapper=mapper)

        weights_util = aa.util.regularization.adaptive_regularization_weights_from(
            inner_coefficient=10.0, outer_coefficient=15.0, pixel_signals=pixel_signals
        )

        assert (weights == weights_util).all()

    def test__regularization_matrix__compare_to_regularization_util(self):

        reg = aa.reg.AdaptiveBrightness(
            inner_coefficient=1.0, outer_coefficient=2.0, signal_scale=1.0
        )

        pixel_neighbors = np.array(
            [
                [1, 4, -1, -1],
                [2, 4, 0, -1],
                [3, 4, 5, 1],
                [5, 2, -1, -1],
                [5, 0, 1, 2],
                [2, 3, 4, -1],
            ]
        )

        pixel_neighbors_size = np.array([2, 3, 4, 2, 4, 3])
        pixel_signals = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])

        pixelization_grid = MockPixelizationGrid(
            pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size
        )

        mapper = MockRegMapper(
            source_pixelization_grid=pixelization_grid, pixel_signals=pixel_signals
        )

        regularization_matrix = reg.regularization_matrix_from_mapper(mapper=mapper)

        regularization_weights = aa.util.regularization.adaptive_regularization_weights_from(
            pixel_signals=pixel_signals, inner_coefficient=1.0, outer_coefficient=2.0
        )

        regularization_matrix_util = aa.util.regularization.weighted_regularization_matrix_from(
            regularization_weights=regularization_weights,
            pixel_neighbors=pixel_neighbors,
            pixel_neighbors_size=pixel_neighbors_size,
        )

        assert (regularization_matrix == regularization_matrix_util).all()



from autoarray.structures.lines.lines import Line1D
from autoarray.structures.grids.irregular import Grid2DIrregularGrouped
from autoarray.structures.grids.irregular import Grid2DIrregularGroupedUniform
from autoarray.structures.grids.iterate import Grid2DIterate
from autoarray.structures.vector_fields.vector_field_irregular import (
    VectorField2DIrregular,
)
from . import exc
from . import plot
from . import util
from . import mock
from .dataset import preprocess
from .dataset.imaging import Imaging
from .dataset.imaging import MaskedImaging
from .dataset.imaging import SettingsMaskedImaging
from .dataset.imaging import SimulatorImaging
from .dataset.interferometer import Interferometer
from .dataset.interferometer import MaskedInterferometer
from .dataset.interferometer import SettingsMaskedInterferometer
from .dataset.interferometer import SimulatorInterferometer
from .fit.fit import FitImaging
from .fit.fit import FitInterferometer
from .instruments import acs
from .instruments import euclid
from .inversion import pixelizations as pix
from .inversion import regularization as reg
from .inversion.inversions import SettingsInversion
from .inversion.inversions import inversion as Inversion
from .inversion.mappers import mapper as Mapper
from .inversion.pixelizations import SettingsPixelization
from .mask.mask_1d import Mask1D
from .mask.mask_2d import Mask2D
from .mock import mock
from .mock import fixtures
from .operators.convolver import Convolver
from .operators.convolver import Convolver
from .operators.transformer import TransformerDFT
from .operators.transformer import TransformerNUFFT
from .structures.arrays import Array2D
from .structures.arrays import ValuesIrregularGrouped
from .structures.arrays.abstract_array import ExposureInfo
from .structures.frames import Frame2D
from .structures.frames.abstract_frame import Scans
from .structures.grids import Grid2D
from .structures.grids import Grid2DInterpolate
from .structures.grids import Grid2DRectangular
from .structures.grids import Grid2DVoronoi
from .structures.grids import Grid2DIrregular
from .structures.grids import Grid2DIrregularGrouped
from .structures.grids import Grid2DIrregularGroupedUniform
from .structures.region import Region2D
from .structures.kernel import Kernel2D
from .structures.region import Region2D
from .structures.visibilities import Visibilities
from .structures.visibilities import VisibilitiesNoiseMap

from autoconf import conf

conf.instance.register(__file__)

__version__ = '0.17.1'



import autoarray.plot as aplt
import pytest

from os import path

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="plot_path")
def make_plot_path_setup():
    return path.join(
        "{}".format(path.dirname(path.realpath(__file__))),
        "files",
        "plots",
        "fit_interferometer",
    )


def test__fit_quantities_are_output(fit_interferometer_7, plot_path, plot_patch):

    fit_interferometer_plotter = aplt.FitInterferometerPlotter(
        fit=fit_interferometer_7,
        mat_plot_1d=aplt.MatPlot1D(output=aplt.Output(path=plot_path, format="png")),
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(path=plot_path, format="png")),
    )

    fit_interferometer_plotter.figures(
        visibilities=True,
        noise_map=True,
        signal_to_noise_map=True,
        model_visibilities=True,
        residual_map_real=True,
        residual_map_imag=True,
        normalized_residual_map_real=True,
        normalized_residual_map_imag=True,
        chi_squared_map_real=True,
        chi_squared_map_imag=True,
    )

    assert path.join(plot_path, "visibilities.png") in plot_patch.paths
    assert path.join(plot_path, "noise_map.png") in plot_patch.paths
    assert path.join(plot_path, "signal_to_noise_map.png") in plot_patch.paths
    assert path.join(plot_path, "model_visibilities.png") in plot_patch.paths
    assert (
        path.join(plot_path, "real_residual_map_vs_uv_distances.png")
        in plot_patch.paths
    )
    assert (
        path.join(plot_path, "real_residual_map_vs_uv_distances.png")
        in plot_patch.paths
    )
    assert (
        path.join(plot_path, "real_normalized_residual_map_vs_uv_distances.png")
        in plot_patch.paths
    )
    assert (
        path.join(plot_path, "imag_normalized_residual_map_vs_uv_distances.png")
        in plot_patch.paths
    )
    assert (
        path.join(plot_path, "imag_chi_squared_map_vs_uv_distances.png")
        in plot_patch.paths
    )
    assert (
        path.join(plot_path, "imag_chi_squared_map_vs_uv_distances.png")
        in plot_patch.paths
    )

    plot_patch.paths = []

    fit_interferometer_plotter.figures(
        visibilities=True,
        noise_map=False,
        signal_to_noise_map=False,
        model_visibilities=True,
        chi_squared_map_real=True,
        chi_squared_map_imag=True,
    )

    assert path.join(plot_path, "visibilities.png") in plot_patch.paths
    assert path.join(plot_path, "noise_map.png") not in plot_patch.paths
    assert path.join(plot_path, "signal_to_noise_map.png") not in plot_patch.paths
    assert path.join(plot_path, "model_visibilities.png") in plot_patch.paths
    assert (
        path.join(plot_path, "real_residual_map_vs_uv_distances.png")
        not in plot_patch.paths
    )
    assert (
        path.join(plot_path, "imag_residual_map_vs_uv_distances.png")
        not in plot_patch.paths
    )
    assert (
        path.join(plot_path, "real_normalized_residual_map_vs_uv_distances.png")
        not in plot_patch.paths
    )
    assert (
        path.join(plot_path, "imag_normalized_residual_map_vs_uv_distances.png")
        not in plot_patch.paths
    )
    assert (
        path.join(plot_path, "real_chi_squared_map_vs_uv_distances.png")
        in plot_patch.paths
    )
    assert (
        path.join(plot_path, "imag_chi_squared_map_vs_uv_distances.png")
        in plot_patch.paths
    )


def test__fit_sub_plot(fit_interferometer_7, plot_path, plot_patch):

    fit_interferometer_plotter = aplt.FitInterferometerPlotter(
        fit=fit_interferometer_7,
        mat_plot_1d=aplt.MatPlot1D(output=aplt.Output(path=plot_path, format="png")),
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(path=plot_path, format="png")),
    )

    fit_interferometer_plotter.subplot_fit_interferometer()

    assert path.join(plot_path, "subplot_fit_interferometer.png") in plot_patch.paths



import inspect
import os

from autoconf import conf
from autoarray import decorator_util
from autoarray.util import mask_util
import numpy as np
from astropy.io import fits
from functools import wraps


class Memoizer:
    def __init__(self):
        """
        Class to store the results of a function given a set of inputs.
        """
        self.results = {}
        self.calls = 0
        self.arg_names = None

    def __call__(self, func):
        """
        Memoize decorator. Any time a function is called that a memoizer has been attached to its results are stored in
        the results dictionary or retrieved from the dictionary if the function has aaready been called with those
        arguments.

        Note that the same memoizer persists over all instances of a class. Any state for a given instance that is not
        given in the representation of that instance will be ignored. That is, it is possible that the memoizer will
        give incorrect results if instance state does not affect __str__ but does affect the value returned by the
        memoized method.

        Parameters
        ----------
        func: function
            A function for which results should be memoized

        Returns
        -------
        decorated : function
            A function that memoizes results
        """
        if self.arg_names is not None:
            raise AssertionError("Instantiate a new Memoizer for each function")
        self.arg_names = inspect.getfullargspec(func).args

        @wraps(func)
        def wrapper(*args, **kwargs):
            key = ", ".join(
                [
                    "('{}', {})".format(arg_name, arg)
                    for arg_name, arg in list(zip(self.arg_names, args))
                    + [(k, v) for k, v in kwargs.items()]
                ]
            )
            if key not in self.results:
                self.calls += 1
            self.results[key] = func(*args, **kwargs)
            return self.results[key]

        return wrapper


@decorator_util.jit()
def extracted_array_2d_from(
    array_2d: np.ndarray, y0: int, y1: int, x0: int, x1: int
) -> np.ndarray:
    """
    Resize an array to a new size by extracting a sub-set of the array.

    The extracted input coordinates use NumPy convention, such that the upper values should be specified as +1 the
    dimensions of the extracted array.

    In the example below, an array of size (5,5) is extracted using the coordinates y0=1, y1=4, x0=1, x1=4. This
    extracts an array of dimensions (3,3) and is equivalent to array_2d[1:4, 1:4].

    This function is necessary work with numba jit tags and is why a standard Numpy array extraction is not used.

    Parameters
    ----------
    array_2d : np.ndarray
        The 2D array that an array is extracted from.
    y0 : int
        The top row number (e.g. the higher y-coodinate) of the array that is extracted for the resize.
    y1 : int
        The bottom row number (e.g. the lower y-coodinate) of the array that is extracted for the resize.
    x0 : int
        The left column number (e.g. the lower x-coodinate) of the array that is extracted for the resize.
    x1 : int
        The right column number (e.g. the higher x-coodinate) of the array that is extracted for the resize.

    Returns
    -------
    ndarray
        The extracted 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    extracted_array = extract_array_2d(array_2d=array_2d, y0=1, y1=4, x0=1, x1=4)
    """

    new_shape = (y1 - y0, x1 - x0)

    resized_array = np.zeros(shape=new_shape)

    for y_resized, y in enumerate(range(y0, y1)):
        for x_resized, x in enumerate(range(x0, x1)):
            if (
                y >= 0
                and x >= 0
                and y <= array_2d.shape[0] - 1
                and x <= array_2d.shape[1] - 1
            ):
                resized_array[y_resized, x_resized] = array_2d[y, x]

    return resized_array


@decorator_util.jit()
def resized_array_2d_from_array_2d(
    array_2d: np.ndarray,
    resized_shape: (int, int),
    origin: (int, int) = (-1, -1),
    pad_value: int = 0.0,
) -> np.ndarray:
    """
    Resize an array to a new size around a central pixel.

    If the origin (e.g. the central pixel) of the resized array is not specified, the central pixel of the array is
    calculated automatically. For example, a (5,5) array's central pixel is (2,2). For even dimensions the central
    pixel is assumed to be the lower indexed value, e.g. a (6,4) array's central pixel is calculated as (2,1).

    The default origin is (-1, -1) because numba requires that the function input is the same type throughout the
    function, thus a default 'None' value cannot be used.

    Parameters
    ----------
    array_2d : np.ndarray
        The 2D array that is resized.
    resized_shape : (int, int)
        The (y,x) new pixel dimension of the trimmed array.
    origin : (int, int)
        The oigin of the resized array, e.g. the central pixel around which the array is extracted.
    pad_value : float
        If the reszied array is bigger in size than the input array, the value the padded edge values are filled in
        using.

    Returns
    -------
    ndarray
        The resized 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = resize_array_2d(array_2d=array_2d, new_shape=(2,2), origin=(2, 2))
    """

    y_is_even = int(array_2d.shape[0]) % 2 == 0
    x_is_even = int(array_2d.shape[1]) % 2 == 0

    if origin == (-1, -1):

        if y_is_even:
            y_centre = int(array_2d.shape[0] / 2)
        elif not y_is_even:
            y_centre = int(array_2d.shape[0] / 2)

        if x_is_even:
            x_centre = int(array_2d.shape[1] / 2)
        elif not x_is_even:
            x_centre = int(array_2d.shape[1] / 2)

        origin = (y_centre, x_centre)

    resized_array = np.zeros(shape=resized_shape)

    if y_is_even:
        y_min = origin[0] - int(resized_shape[0] / 2)
        y_max = origin[0] + int((resized_shape[0] / 2)) + 1
    elif not y_is_even:
        y_min = origin[0] - int(resized_shape[0] / 2)
        y_max = origin[0] + int((resized_shape[0] / 2)) + 1

    if x_is_even:
        x_min = origin[1] - int(resized_shape[1] / 2)
        x_max = origin[1] + int((resized_shape[1] / 2)) + 1
    elif not x_is_even:
        x_min = origin[1] - int(resized_shape[1] / 2)
        x_max = origin[1] + int((resized_shape[1] / 2)) + 1

    for y_resized, y in enumerate(range(y_min, y_max)):
        for x_resized, x in enumerate(range(x_min, x_max)):
            if y >= 0 and y < array_2d.shape[0] and x >= 0 and x < array_2d.shape[1]:
                if (
                    y_resized >= 0
                    and y_resized < resized_shape[0]
                    and x_resized >= 0
                    and x_resized < resized_shape[1]
                ):
                    resized_array[y_resized, x_resized] = array_2d[y, x]
            else:
                if (
                    y_resized >= 0
                    and y_resized < resized_shape[0]
                    and x_resized >= 0
                    and x_resized < resized_shape[1]
                ):
                    resized_array[y_resized, x_resized] = pad_value

    return resized_array


@decorator_util.jit()
def replace_noise_map_2d_values_where_image_2d_values_are_negative(
    image_2d: np.ndarray, noise_map_2d: np.ndarray, target_signal_to_noise: float = 2.0
) -> np.ndarray:
    """
    If the values of a 2D image array are negative, this function replaces the corresponding 2D noise-map array
    values to meet a specified target to noise value.

    This routine is necessary because of anomolous values in images which come from our HST ACS data_type-reduction
    pipeline, where image-pixels with negative values (e.g. due to the background sky subtraction) have extremely
    small noise values, which inflate their signal-to-noise values and chi-squared contributions in the modeling.

    Parameters
    ----------
    image_2d : np.ndarray
        The 2D image array used to locate the pixel indexes in the noise-map which are replaced.
    noise_map_2d : np.ndarray
        The 2D noise-map array whose values are replaced.
    target_signal_to_noise : float
        The target signal-to-noise the noise-map valueus are changed to.

    Returns
    -------
    ndarray
        The 2D noise-map with values changed.

    Examples
    --------
    image_2d = np.ones((5,5))
    image_2d[2,2] = -1.0
    noise_map_2d = np.ones((5,5))

    noise_map_2d_replaced = replace_noise_map_2d_values_where_image_2d_values_are_negative(
        image_2d=image_2d, noise_map_2d=noise_map_2d, target_signal_to_noise=2.0):
    """
    for y in range(image_2d.shape[0]):
        for x in range(image_2d.shape[1]):
            if image_2d[y, x] < 0.0:
                absolute_signal_to_noise = np.abs(image_2d[y, x]) / noise_map_2d[y, x]
                if absolute_signal_to_noise >= target_signal_to_noise:
                    noise_map_2d[y, x] = np.abs(image_2d[y, x]) / target_signal_to_noise

    return noise_map_2d


def numpy_array_1d_to_fits(
    array_1d: np.ndarray, file_path: str, overwrite: bool = False
):
    """
    Write a 1D NumPy array to a .fits file.

    Parameters
    ----------
    array_2d : np.ndarray
        The 2D array that is written to fits.
    file_path : str
        The full path of the file that is output, including the file name and ``.fits`` extension.
    overwrite : bool
        If `True` and a file already exists with the input file_path the .fits file is overwritten. If False, an error
        will be raised.

    Returns
    -------
    None

    Examples
    --------
    array_2d = np.ones((5,5))
    numpy_array_to_fits(array_2d=array_2d, file_path='/path/to/file/filename.fits', overwrite=True)
    """

    file_dir = os.path.split(file_path)[0]

    if not os.path.exists(file_dir):
        os.makedirs(file_dir)

    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    new_hdr = fits.Header()
    hdu = fits.PrimaryHDU(array_1d, new_hdr)
    hdu.writeto(file_path)


def numpy_array_1d_from_fits(file_path: str, hdu: int):
    """
    Read a 1D NumPy array from a .fits file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the structures
    appear the same orientation as .fits files loaded in DS9.

    Parameters
    ----------
    file_path : str
        The full path of the file that is loaded, including the file name and ``.fits`` extension.
    hdu : int
        The HDU extension of the array that is loaded from the .fits file.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .fits file.

    Examples
    --------
    array_2d = numpy_array_from_fits(file_path='/path/to/file/filename.fits', hdu=0)
    """
    hdu_list = fits.open(file_path)
    return np.array(hdu_list[hdu].data)


def numpy_array_2d_to_fits(
    array_2d: np.ndarray, file_path: str, overwrite: bool = False
):
    """
    Write a 2D NumPy array to a .fits file.

    Before outputting a NumPy array, the array may be flipped upside-down using np.flipud depending on the project
    config files. This is for Astronomy projects so that structures appear the same orientation as ``.fits`` files
    loaded in DS9.

    Parameters
    ----------
    array_2d : np.ndarray
        The 2D array that is written to fits.
    file_path : str
        The full path of the file that is output, including the file name and ``.fits`` extension.
    overwrite : bool
        If `True` and a file already exists with the input file_path the .fits file is overwritten. If `False`, an
        error is raised.

    Returns
    -------
    None

    Examples
    --------
    array_2d = np.ones((5,5))
    numpy_array_to_fits(array_2d=array_2d, file_path='/path/to/file/filename.fits', overwrite=True)
    """

    file_dir = os.path.split(file_path)[0]

    if not os.path.exists(file_dir):
        os.makedirs(file_dir)

    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    new_hdr = fits.Header()

    flip_for_ds9 = conf.instance["general"]["fits"]["flip_for_ds9"]

    if flip_for_ds9:
        hdu = fits.PrimaryHDU(np.flipud(array_2d), new_hdr)
    else:
        hdu = fits.PrimaryHDU(array_2d, new_hdr)
    hdu.writeto(file_path)


def numpy_array_2d_from_fits(
    file_path: str, hdu: int, do_not_scale_image_data: bool = False
):
    """
    Read a 2D NumPy array from a .fits file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the structures
    appear the same orientation as .fits files loaded in DS9.

    Parameters
    ----------
    file_path : str
        The full path of the file that is loaded, including the file name and ``.fits`` extension.
    hdu : int
        The HDU extension of the array that is loaded from the .fits file.
    do_not_scale_image_data : bool
        If True, the .fits file is not rescaled automatically based on the .fits header info.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .fits file.

    Examples
    --------
    array_2d = numpy_array_from_fits(file_path='/path/to/file/filename.fits', hdu=0)
    """
    hdu_list = fits.open(file_path, do_not_scale_image_data=do_not_scale_image_data)

    flip_for_ds9 = conf.instance["general"]["fits"]["flip_for_ds9"]

    if flip_for_ds9:
        return np.flipud(np.array(hdu_list[hdu].data)).astype("float64")
    return np.array(hdu_list[hdu].data).astype("float64")


@decorator_util.jit()
def index_2d_for_index_slim_from(indexes_slim: np.ndarray, shape_native) -> np.ndarray:
    """
    For pixels on a native 2D array of shape (total_y_pixels, total_x_pixels), this array maps the slimmed 1D pixel
    indexes to their corresponding 2D pixel indexes.

    Indexing is defined from the top-left corner rightwards and downwards, whereby the top-left pixel on the 2D array
    corresponds to index 0, the pixel to its right pixel 1, and so on.

    For a 2D array of shape (3,3), flattened 1D pixel indexes are converted as follows:

    - 1D Pixel index 0 maps -> 2D pixel index [0,0].
    - 1D Pixel index 1 maps -> 2D pixel index [0,1].
    - 1D Pixel index 4 maps -> 2D pixel index [1,0].
    - 1D Pixel index 8 maps -> 2D pixel index [2,2].

    Parameters
    ----------
    indexes_slim : np.ndarray
        The slim 1D pixel indexes which are mapped to 2D indexes.
    shape_native : (int, int)
        The shape of the 2D array which the pixels are natively defined on.

    Returns
    -------
    ndarray
        An array of native 2d pixel indexes with dimensions (total_indexes, 2).

    Examples
    --------
    indexes_slim = np.array([0, 1, 2, 5])
    indexes_2d = index_2d_for_index_slim_from(indexes_slim=indexes_slim, shape=(3,3))
    """
    index_2d_for_index_slim = np.zeros((indexes_slim.shape[0], 2))

    for i, index_slim in enumerate(indexes_slim):

        index_2d_for_index_slim[i, 0] = int(index_slim / shape_native[1])
        index_2d_for_index_slim[i, 1] = int(index_slim % shape_native[1])

    return index_2d_for_index_slim


@decorator_util.jit()
def index_slim_for_index_2d_from(indexes_2d: np.ndarray, shape_native) -> np.ndarray:
    """
    For pixels on a native 2D array of shape (total_y_pixels, total_x_pixels), this array maps the 2D pixel indexes to
    their corresponding slimmed 1D pixel indexes.

    Indexing is defined from the top-left corner rightwards and downwards, whereby the top-left pixel on the 2D array
    corresponds to index 0, the pixel to its right pixel 1, and so on.

    For a 2D array of shape (3,3), 2D pixel indexes are converted as follows:

    - 2D Pixel index [0,0] maps -> 1D pixel index 0.
    - 2D Pixel index [0,1] maps -> 2D pixel index 1.
    - 2D Pixel index [1,0] maps -> 2D pixel index 4.
    - 2D Pixel index [2,2] maps -> 2D pixel index 8.

    Parameters
    ----------
    indexes_2d : np.ndarray
        The native 2D pixel indexes which are mapped to slimmed 1D indexes.
    shape_native : (int, int)
        The shape of the 2D array which the pixels are defined on.

    Returns
    -------
    ndarray
        An array of 1d pixel indexes with dimensions (total_indexes).

    Examples
    --------
    indexes_2d = np.array([[0,0], [1,0], [2,0], [2,2]])
    indexes_flat = index_flat_for_index_2d_from(indexes_2d=indexes_2d, shape=(3,3))
    """
    index_slim_for_index_native_2d = np.zeros(indexes_2d.shape[0])

    for i in range(indexes_2d.shape[0]):
        index_slim_for_index_native_2d[i] = int(
            (indexes_2d[i, 0]) * shape_native[1] + indexes_2d[i, 1]
        )

    return index_slim_for_index_native_2d


@decorator_util.jit()
def sub_array_2d_slim_from(
    sub_array_2d: np.ndarray, mask_2d: np.ndarray, sub_size: int
) -> np.ndarray:
    """
    For a 2D sub array and mask, map the values of all unmasked pixels to its slimmed 1D sub-array.

    A sub-array is an array whose native dimensions correspond to the normal array multiplied by the sub_size. For
    example if an array is shape [total_y_pixels, total_x_pixels] and the ``sub_size=2``, the sub_array is shape
    [total_y_pixels*sub_size, total_x_pixels*sub_size].

    The pixel coordinate origin is at the top left corner of the 2D array and goes right-wards and downwards,
    with sub-pixels then going right and downwards in each pixel. For example, for an array of shape (3,3) and a
    sub-grid size of 2 where all pixels are unmasked:

    - pixel [0,0] of the 2D array will correspond to index 0 of the 1D array.
    - pixel [0,1] of the 2D array will correspond to index 1 of the 1D array.
    - pixel [1,0] of the 2D array will correspond to index 2 of the 1D array.
    - pixel [2,0] of the 2D array will correspond to index 4 of the 1D array.
    - pixel [1,0] of the 2D array will correspond to index 12 of the 1D array.

    Parameters
    ----------
    sub_array_2d : np.ndarray
        A 2D array of values on the dimensions of the sub-grid.
    mask_2d : np.ndarray
        A 2D array of bools, where `False` values mean unmasked and are included in the mapping.
    array_2d : np.ndarray
        The 2D array of values which are mapped to a 1D array.

    Returns
    -------
    ndarray
        The slimmed 1D array of values mapped from the native 2D array with dimensions (total_unmasked_pixels).

    Examples
    --------

    sub_array_2d = np.array([[ 1.0,  2.0,  5.0,  6.0],
                             [ 3.0,  4.0,  7.0,  8.0],
                             [ 9.0, 10.0, 13.0, 14.0],
                             [11.0, 12.0, 15.0, 16.0])

    mask = np.array([[True, False],
                     [False, False]])

    sub_array_2d_slim = sub_array_2d_slim_from(mask=mask, array_2d=array_2d, sub_size=2)
    """

    total_sub_pixels = mask_util.total_sub_pixels_2d_from(
        mask_2d=mask_2d, sub_size=sub_size
    )

    sub_array_2d_slim = np.zeros(shape=total_sub_pixels)
    index = 0

    for y in range(mask_2d.shape[0]):
        for x in range(mask_2d.shape[1]):
            if not mask_2d[y, x]:
                for y1 in range(sub_size):
                    for x1 in range(sub_size):
                        sub_array_2d_slim[index] = sub_array_2d[
                            y * sub_size + y1, x * sub_size + x1
                        ]
                        index += 1

    return sub_array_2d_slim


def sub_array_2d_from(
    sub_array_2d_slim: np.ndarray, mask_2d: np.ndarray, sub_size: int
) -> np.ndarray:
    """
    For a slimmed 2D array that was computed by mapping unmasked values from a native 2D array of shape
    [total_y_pixels, total_x_pixels], map its values back to the original 2D array where masked values are set to zero.

    This uses the array ``slim_to_native`` where each index gives the 2D sub pixel indexes of the 1D array's
    unmasked pixels, for example:

    - If ``slim_to_native[0] = [0,0]``, the first value of the 1D array maps to the pixel [0,0] of the 2D array.
    - If ``slim_to_native[1] = [0,1]``, the second value of the 1D array maps to the pixel [0,1] of the 2D array.
    - If ``slim_to_native[4] = [1,1]``, the fifth value of the 1D array maps to the pixel [1,1] of the 2D array.

    This example uses an array `sub_slim_to_native` which includes sub-gridding in the mapping.

    Parameters
    ----------
    sub_array_2d_slim : np.ndarray
        The slimmed array of values which are mapped to a 2D array.
    mask_2d : np.ndarray
        A 2D array of bools, where `False` values mean unmasked and are included in the mapping.
    sub_size : int
        The size (sub_size x sub_size) of each unmasked pixels sub-array.

    Returns
    -------
    ndarray
        A 2D array of values mapped from the 1D array with dimensions shape.

    Examples
    --------
    slim_to_native = np.array([[0,1], [1,0], [1,1], [1,2], [2,1]])

    array_1d = np.array([[2.0, 4.0, 5.0, 6.0, 8.0])

    array_2d = map_masked_1d_array_to_2d_array_from_array_1d_shape_and_slim_to_native(
        array_1d=array_1d, shape=(3,3), slim_to_native=slim_to_native)
    """

    sub_shape = (mask_2d.shape[0] * sub_size, mask_2d.shape[1] * sub_size)

    sub_slim_to_native = mask_util.sub_native_index_for_sub_slim_index_via_mask_2d_from(
        mask_2d=mask_2d, sub_size=sub_size
    ).astype("int")

    return sub_array_2d_via_sub_indexes_from(
        sub_array_2d_slim=sub_array_2d_slim,
        sub_shape=sub_shape,
        sub_native_index_for_slim_index=sub_slim_to_native,
    )


@decorator_util.jit()
def sub_array_2d_via_sub_indexes_from(
    sub_array_2d_slim: np.ndarray,
    sub_shape: (int, int),
    sub_native_index_for_slim_index: np.ndarray,
) -> np.ndarray:
    """
    For a slimmed sub array and sub-indexes mapping the slimmed sub-array values to their native sub-array,
    return the native 2D sub-array.

    A sub-array is an array whose dimensions correspond to the normal array multiplied by the sub_size. For example
    if an array is shape [total_y_pixels, total_x_pixels] and the ``sub_size=2``, the sub_array is shape
    [total_y_pixels*sub_size, total_x_pixels*sub_size].

    The pixel coordinate origin is at the top left corner of the 2D array and goes right-wards and downwards,
    with sub-pixels then going right and downwards in each pixel. For example, for an array of shape (3,3) and a
    sub-grid size of 2 where all pixels are unmasked:

    - pixel [0,0] of the native 2D array will correspond to index 0 of the slim array.
    - pixel [0,1] of the native 2D array will correspond to index 1 of the slim array.
    - pixel [1,0] of the native 2D array will correspond to index 2 of the slim array.
    - pixel [2,0] of the native 2D array will correspond to index 4 of the slim array.
    - pixel [1,0] of the native 2D array will correspond to index 12 of the slim array.

    Parameters
    ----------
    sub_array_2d_slim : np.ndarray
        The slimmed array of shape [total_values] which are mapped to the native array..
    sub_shape : (int, int)
        The 2D dimensions of the native 2D sub-array.
    sub_native_index_for_slim_index : np.narray
        An array of shape [total_values] that maps from the slimmed array to the native array.

    Returns
    -------
    ndarray
        The native 2D array of values mapped from the slimmed array with dimensions (total_values, total_values).
    """
    sub_array_native_2d = np.zeros(sub_shape)

    for slim_index in range(len(sub_native_index_for_slim_index)):
        sub_array_native_2d[
            sub_native_index_for_slim_index[slim_index, 0],
            sub_native_index_for_slim_index[slim_index, 1],
        ] = sub_array_2d_slim[slim_index]

    return sub_array_native_2d


@decorator_util.jit()
def sub_array_complex_slim_from(
    sub_array_2d: np.ndarray, mask: np.ndarray, sub_size: int
) -> np.ndarray:
    """
    For a 2D sub array and mask, map the values of all unmasked pixels to a 1D sub-array.

    A sub-array is an array whose dimensions correspond to the normal array (e.g. used to make the grid)
    multiplied by the sub_size. E.g., it is an array that would be generated using the sub-grid and not binning
    up values in sub-pixels back to the grid.

    The pixel coordinate origin is at the top left corner of the 2D array and goes right-wards and downwards,
    with sub-pixels then going right and downwards in each pixel. For example, for an array of shape (3,3) and a
    sub-grid size of 2 where all pixels are unmasked:

    - pixel [0,0] of the 2D array will correspond to index 0 of the 1D array.
    - pixel [0,1] of the 2D array will correspond to index 1 of the 1D array.
    - pixel [1,0] of the 2D array will correspond to index 2 of the 1D array.
    - pixel [2,0] of the 2D array will correspond to index 4 of the 1D array.
    - pixel [1,0] of the 2D array will correspond to index 12 of the 1D array.

    Parameters
    ----------
    sub_array_2d : np.ndarray
        A 2D array of values on the dimensions of the sub-grid.
    mask : np.ndarray
        A 2D array of bools, where `False` values mean unmasked and are included in the mapping.
    array_2d : np.ndarray
        The 2D array of values which are mapped to a 1D array.

    Returns
    -------
    ndarray
        A 1D array of values mapped from the 2D array with dimensions (total_unmasked_pixels).

    Examples
    --------

    sub_array_2d = np.array([[ 1.0,  2.0,  5.0,  6.0],
                             [ 3.0,  4.0,  7.0,  8.0],
                             [ 9.0, 10.0, 13.0, 14.0],
                             [11.0, 12.0, 15.0, 16.0])

    mask = np.array([[True, False],
                     [False, False]])

    sub_array_1d = map_sub_array_2d_to_masked_sub_array_1d_from_sub_array_2d_mask_and_sub_size(
        mask=mask, array_2d=array_2d)
    """

    total_sub_pixels = mask_util.total_sub_pixels_2d_from(
        mask_2d=mask, sub_size=sub_size
    )

    sub_array_1d = 0 + 0j * np.zeros(shape=total_sub_pixels)
    index = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                for y1 in range(sub_size):
                    for x1 in range(sub_size):
                        sub_array_1d[index] = sub_array_2d[
                            y * sub_size + y1, x * sub_size + x1
                        ]
                        index += 1

    return sub_array_1d


@decorator_util.jit()
def sub_array_2d_complex_via_sub_indexes_from(
    sub_array_2d_slim: np.ndarray,
    sub_shape_native: (int, int),
    sub_native_index_for_slim_index: np.ndarray,
) -> np.ndarray:

    sub_array_2d = 0 + 0j * np.zeros(sub_shape_native)

    for slim_index in range(len(sub_native_index_for_slim_index)):
        sub_array_2d[
            sub_native_index_for_slim_index[slim_index, 0],
            sub_native_index_for_slim_index[slim_index, 1],
        ] = sub_array_2d_slim[slim_index]

    return sub_array_2d



import numpy as np
import pytest

import autoarray as aa
from autoarray.structures import grids

from autoarray.mock.mock import ndarray_1d_from_grid, ndarray_2d_from_grid


class TestObj:
    def test__blurring_grid_from_mask__compare_to_array_util(self):
        mask = np.array(
            [
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, False, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
            ]
        )

        mask = aa.Mask2D.manual(mask=mask, pixel_scales=(2.0, 2.0), sub_size=2)

        blurring_mask_util = aa.util.mask.blurring_mask_2d_from(
            mask_2d=mask, kernel_shape_native=(3, 5)
        )

        blurring_grid_util = aa.util.grid.grid_2d_slim_via_mask_from(
            mask_2d=blurring_mask_util, pixel_scales=(2.0, 2.0), sub_size=1
        )

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.1)

        blurring_grid = grid.blurring_grid_from_kernel_shape(kernel_shape_native=(3, 5))

        assert isinstance(blurring_grid, grids.Grid2DInterpolate)
        assert len(blurring_grid.shape) == 2
        assert blurring_grid == pytest.approx(blurring_grid_util, 1e-4)
        assert blurring_grid.pixel_scales == (2.0, 2.0)
        assert blurring_grid.pixel_scales_interp == (0.1, 0.1)

    def test__blurring_grid_from_kernel_shape__compare_to_array_util(self):
        mask = np.array(
            [
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, False, True, True, True, False, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, False, True, True, True, False, True, True],
                [True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True],
            ]
        )

        mask = aa.Mask2D.manual(mask=mask, pixel_scales=(2.0, 2.0))

        blurring_mask_util = aa.util.mask.blurring_mask_2d_from(
            mask_2d=mask, kernel_shape_native=(3, 5)
        )

        blurring_grid_util = aa.util.grid.grid_2d_slim_via_mask_from(
            mask_2d=blurring_mask_util, pixel_scales=(2.0, 2.0), sub_size=1
        )

        mask = aa.Mask2D.manual(mask=mask, pixel_scales=(2.0, 2.0))

        blurring_grid = grids.Grid2DInterpolate.blurring_grid_from_mask_and_kernel_shape(
            mask=mask, kernel_shape_native=(3, 5), pixel_scales_interp=0.1
        )

        assert isinstance(blurring_grid, grids.Grid2DInterpolate)
        assert len(blurring_grid.shape) == 2
        assert blurring_grid == pytest.approx(blurring_grid_util, 1e-4)
        assert blurring_grid.pixel_scales == (2.0, 2.0)
        assert blurring_grid.pixel_scales_interp == (0.1, 0.1)

        blurring_grid = grids.Grid2DInterpolate.blurring_grid_from_mask_and_kernel_shape(
            mask=mask,
            kernel_shape_native=(3, 5),
            pixel_scales_interp=0.1,
            store_slim=False,
        )

        assert isinstance(blurring_grid, grids.Grid2DInterpolate)
        assert len(blurring_grid.shape) == 3
        assert blurring_grid.pixel_scales == (2.0, 2.0)
        assert blurring_grid.pixel_scales_interp == (0.1, 0.1)

    def test__padded_grid_from_kernel_shape(self):
        grid = grids.Grid2DInterpolate.uniform(
            shape_native=(4, 4), pixel_scales=3.0, pixel_scales_interp=0.1
        )

        padded_grid = grid.padded_grid_from_kernel_shape(kernel_shape_native=(3, 3))

        assert isinstance(padded_grid, grids.Grid2DInterpolate)
        assert padded_grid.pixel_scales_interp == (0.1, 0.1)

        mask = aa.Mask2D.unmasked(
            shape_native=(6, 6), pixel_scales=(3.0, 3.0), sub_size=1
        )

        grid = grids.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.1)

        assert isinstance(padded_grid, grids.Grid2DInterpolate)
        assert padded_grid.pixel_scales_interp == (0.1, 0.1)
        assert (padded_grid.vtx == grid.vtx).all()
        assert (padded_grid.wts == grid.wts).all()

        mask = aa.Mask2D.manual(
            mask=np.full((2, 5), False), pixel_scales=(8.0, 8.0), sub_size=4
        )

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.2)

        padded_grid = grid.padded_grid_from_kernel_shape(kernel_shape_native=(5, 5))

        padded_grid_util = aa.util.grid.grid_2d_slim_via_mask_from(
            mask_2d=np.full((6, 9), False), pixel_scales=(8.0, 8.0), sub_size=4
        )

        assert isinstance(padded_grid, grids.Grid2DInterpolate)
        assert padded_grid.pixel_scales_interp == (0.2, 0.2)
        assert isinstance(padded_grid, grids.Grid2DInterpolate)
        assert padded_grid.shape == (864, 2)
        assert (padded_grid.mask == np.full(fill_value=False, shape=(6, 9))).all()
        assert padded_grid == pytest.approx(padded_grid_util, 1e-4)


class TestInterpolatedResult:
    def test__function_returns_binary_ndarray_1d__returns_interpolated_array(self):

        # noinspection PyUnusedLocal

        class MockInterpolateClass:
            def func(self, profile, grid):
                result = np.zeros(grid.shape[0])
                result[0] = 1
                return result

        mask = aa.Mask2D.unmasked(
            shape_native=(3, 3), pixel_scales=(1.0, 1.0), sub_size=1
        )

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.5)

        cls = MockInterpolateClass()

        interp_array = grid.result_from_func(func=cls.func, cls=MockInterpolateClass())

        assert isinstance(interp_array, aa.Array2D)
        assert interp_array.ndim == 1
        assert interp_array.shape == (9,)
        print(interp_array)
        assert (interp_array != np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0]])).any()

    def test__function_is_false_in_config__does_not_use_interpolatin(self):

        # noinspection PyUnusedLocal

        class MockInterpolateClass:
            def func_off(self, profile, grid):
                result = np.zeros(grid.shape[0])
                result[0] = 1
                return result

        mask = aa.Mask2D.unmasked(
            shape_native=(3, 3), pixel_scales=(1.0, 1.0), sub_size=1
        )

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.5)

        cls = MockInterpolateClass()

        arr = grid.result_from_func(func=cls.func_off, cls=MockInterpolateClass())

        assert isinstance(arr, aa.Array2D)
        assert arr.ndim == 1
        assert arr.shape == (9,)
        assert (arr == np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0]])).any()

    def test__function_returns_binary_ndarray_2d__returns_interpolated_grid(self):

        # noinspection PyUnusedLocal
        class MockInterpolateClass:
            def func(self, profile, grid):
                result = np.zeros((grid.shape[0], 2))
                result[0, :] = 1
                return result

        mask = aa.Mask2D.unmasked(
            shape_native=(3, 3), pixel_scales=(1.0, 1.0), sub_size=1
        )

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.5)

        cls = MockInterpolateClass()

        interp_grid = grid.result_from_func(func=cls.func, cls=MockInterpolateClass())

        assert isinstance(interp_grid, aa.Grid2D)
        assert interp_grid.ndim == 2
        assert interp_grid.shape == (9, 2)
        assert (
            interp_grid
            != np.array(
                np.array(
                    [
                        [1, 1],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                        [0, 0],
                    ]
                )
            )
        ).any()

    def test__function_returns_ndarray_1d__interpolation_used_and_accurate(self):

        # noinspection PyUnusedLocal
        class MockInterpolateObj:
            def ndarray_1d_from_grid(self, profile, grid):
                return ndarray_1d_from_grid(profile=profile, grid=grid)

        cls = MockInterpolateObj()

        mask = aa.Mask2D.circular_annular(
            shape_native=(20, 20),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
            inner_radius=3.0,
            outer_radius=8.0,
        )

        grid = aa.Grid2D.from_mask(mask=mask)

        true_array = ndarray_1d_from_grid(profile=None, grid=grid)

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=1.0)

        interpolated_array = grid.result_from_func(
            func=cls.ndarray_1d_from_grid, cls=MockInterpolateObj()
        )

        assert interpolated_array.shape[0] == mask.pixels_in_mask
        assert (true_array == interpolated_array).all()

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.1)

        interpolated_array = grid.result_from_func(
            func=cls.ndarray_1d_from_grid, cls=MockInterpolateObj()
        )

        assert interpolated_array.shape[0] == mask.pixels_in_mask
        assert true_array[0] != interpolated_array[0]
        assert np.max(true_array - interpolated_array) < 0.001

        mask = aa.Mask2D.circular_annular(
            shape_native=(28, 28),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
            inner_radius=3.0,
            outer_radius=8.0,
            centre=(3.0, 3.0),
        )

        grid = aa.Grid2D.from_mask(mask=mask)

        true_array = ndarray_1d_from_grid(profile=None, grid=grid)

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.1)

        interpolated_array = grid.result_from_func(
            func=cls.ndarray_1d_from_grid, cls=MockInterpolateObj()
        )

        assert interpolated_array.shape[0] == mask.pixels_in_mask
        assert true_array[0] != interpolated_array[0]
        assert np.max(true_array - interpolated_array) < 0.001

    def test__function_returns_ndarray_2d__interpolation_used_and_accurate(self):

        # noinspection PyUnusedLocal
        class MockInterpolateObj:
            def ndarray_2d_from_grid(self, profile, grid):
                return ndarray_2d_from_grid(profile=profile, grid=grid)

        cls = MockInterpolateObj()

        mask = aa.Mask2D.circular_annular(
            shape_native=(20, 20),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
            inner_radius=3.0,
            outer_radius=8.0,
        )

        grid = aa.Grid2D.from_mask(mask=mask)

        true_grid = ndarray_2d_from_grid(profile=None, grid=grid)

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=1.0)

        interpolated_grid = grid.result_from_func(
            func=cls.ndarray_2d_from_grid, cls=MockInterpolateObj()
        )

        assert interpolated_grid.shape[0] == mask.pixels_in_mask
        assert interpolated_grid.shape[1] == 2
        assert (true_grid == interpolated_grid).all()

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.1)

        interpolated_grid = grid.result_from_func(
            func=cls.ndarray_2d_from_grid, cls=MockInterpolateObj()
        )

        assert interpolated_grid.shape[0] == mask.pixels_in_mask
        assert interpolated_grid.shape[1] == 2
        assert true_grid[0, 0] != interpolated_grid[0, 0]
        assert np.max(true_grid[:, 0] - interpolated_grid[:, 0]) < 0.001
        assert np.max(true_grid[:, 1] - interpolated_grid[:, 1]) < 0.001

        mask = aa.Mask2D.circular_annular(
            shape_native=(28, 28),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
            inner_radius=3.0,
            outer_radius=8.0,
            centre=(3.0, 3.0),
        )

        grid = aa.Grid2D.from_mask(mask=mask)

        true_grid = ndarray_2d_from_grid(profile=None, grid=grid)

        grid = aa.Grid2DInterpolate.from_mask(mask=mask, pixel_scales_interp=0.1)

        interpolated_grid = grid.result_from_func(
            func=cls.ndarray_2d_from_grid, cls=MockInterpolateObj()
        )

        assert interpolated_grid.shape[0] == mask.pixels_in_mask
        assert interpolated_grid.shape[1] == 2
        assert true_grid[0, 0] != interpolated_grid[0, 0]
        assert np.max(true_grid[:, 0] - interpolated_grid[:, 0]) < 0.01
        assert np.max(true_grid[:, 1] - interpolated_grid[:, 1]) < 0.01


class TestAPI:
    def test__manual_slim(self):

        grid = aa.Grid2DInterpolate.manual_slim(
            grid=[[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]],
            shape_native=(2, 2),
            pixel_scales=1.0,
            pixel_scales_interp=0.1,
            origin=(0.0, 1.0),
            store_slim=True,
        )

        assert type(grid) == grids.Grid2DInterpolate
        assert type(grid.slim) == grids.Grid2DInterpolate
        assert type(grid.native) == grids.Grid2DInterpolate
        assert (
            grid == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])
        ).all()
        assert (
            grid.native
            == np.array([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
        ).all()
        assert (
            grid.slim == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])
        ).all()
        assert grid.pixel_scales == (1.0, 1.0)
        assert grid.pixel_scales_interp == (0.1, 0.1)
        assert grid.origin == (0.0, 1.0)

        grid = aa.Grid2DInterpolate.manual_slim(
            grid=[[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]],
            shape_native=(2, 2),
            pixel_scales=1.0,
            pixel_scales_interp=0.1,
            origin=(0.0, 1.0),
            store_slim=False,
        )

        assert type(grid) == grids.Grid2DInterpolate
        assert type(grid.slim) == grids.Grid2DInterpolate
        assert type(grid.native) == grids.Grid2DInterpolate
        assert (
            grid == np.array([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
        ).all()
        assert (
            grid.native
            == np.array([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
        ).all()
        assert (
            grid.slim == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])
        ).all()
        assert grid.pixel_scales == (1.0, 1.0)
        assert grid.pixel_scales_interp == (0.1, 0.1)
        assert grid.origin == (0.0, 1.0)

        grid = aa.Grid2DInterpolate.manual_slim(
            grid=[[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]],
            shape_native=(1, 1),
            pixel_scales=1.0,
            pixel_scales_interp=0.1,
            sub_size=2,
            origin=(0.0, 1.0),
            store_slim=True,
        )

        assert type(grid) == grids.Grid2DInterpolate
        assert type(grid.slim) == grids.Grid2DInterpolate
        assert type(grid.native) == grids.Grid2DInterpolate
        assert (
            grid == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])
        ).all()
        assert (
            grid.native
            == np.array([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
        ).all()
        assert (
            grid.slim == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])
        ).all()
        assert (grid.native_binned == np.array([[[4.0, 5.0]]])).all()
        assert (grid.slim_binned == np.array([[4.0, 5.0]])).all()
        assert grid.pixel_scales == (1.0, 1.0)
        assert grid.pixel_scales_interp == (0.1, 0.1)
        assert grid.origin == (0.0, 1.0)
        assert grid.sub_size == 2

    def test__from_mask(self):

        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )
        mask = aa.Mask2D.manual(mask=mask, pixel_scales=(2.0, 2.0), sub_size=1)

        grid_via_util = aa.util.grid.grid_2d_slim_via_mask_from(
            mask_2d=mask, sub_size=1, pixel_scales=(2.0, 2.0)
        )

        grid = aa.Grid2DInterpolate.from_mask(
            mask=mask, pixel_scales_interp=0.1, store_slim=True
        )

        assert type(grid) == grids.Grid2DInterpolate
        assert type(grid.slim) == grids.Grid2DInterpolate
        assert type(grid.native) == grids.Grid2DInterpolate
        assert grid == pytest.approx(grid_via_util, 1e-4)
        assert grid.pixel_scales_interp == (0.1, 0.1)
        assert grid.sub_size == 1

        grid_via_util = aa.util.grid.grid_2d_via_mask_from(
            mask_2d=mask, sub_size=1, pixel_scales=(2.0, 2.0)
        )

        grid = aa.Grid2DInterpolate.from_mask(
            mask=mask, pixel_scales_interp=0.1, store_slim=False
        )

        assert type(grid) == grids.Grid2DInterpolate
        assert type(grid.slim) == grids.Grid2DInterpolate
        assert type(grid.native) == grids.Grid2DInterpolate
        assert grid == pytest.approx(grid_via_util, 1e-4)
        assert grid.pixel_scales_interp == (0.1, 0.1)
        assert grid.sub_size == 1

        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )
        mask = aa.Mask2D.manual(mask=mask, pixel_scales=(2.0, 2.0), sub_size=2)

        grid_via_util = aa.util.grid.grid_2d_slim_via_mask_from(
            mask_2d=mask, sub_size=2, pixel_scales=(2.0, 2.0)
        )

        grid = aa.Grid2DInterpolate.from_mask(
            mask=mask, pixel_scales_interp=0.1, store_slim=True
        )

        assert type(grid) == grids.Grid2DInterpolate
        assert type(grid.slim) == grids.Grid2DInterpolate
        assert type(grid.native) == grids.Grid2DInterpolate
        assert grid == pytest.approx(grid_via_util, 1e-4)
        assert grid.pixel_scales_interp == (0.1, 0.1)
        assert grid.sub_size == 2

    def test__uniform(self):

        grid = aa.Grid2DInterpolate.uniform(
            shape_native=(2, 2),
            pixel_scales=2.0,
            pixel_scales_interp=0.1,
            store_slim=True,
        )

        assert type(grid) == grids.Grid2DInterpolate
        assert type(grid.slim) == grids.Grid2DInterpolate
        assert type(grid.native) == grids.Grid2DInterpolate
        assert (
            grid == np.array([[1.0, -1.0], [1.0, 1.0], [-1.0, -1.0], [-1.0, 1.0]])
        ).all()
        assert (
            grid.native
            == np.array([[[1.0, -1.0], [1.0, 1.0]], [[-1.0, -1.0], [-1.0, 1.0]]])
        ).all()
        assert (
            grid.slim == np.array([[1.0, -1.0], [1.0, 1.0], [-1.0, -1.0], [-1.0, 1.0]])
        ).all()
        assert grid.pixel_scales == (2.0, 2.0)
        assert grid.pixel_scales_interp == (0.1, 0.1)
        assert grid.origin == (0.0, 0.0)

        grid = aa.Grid2DInterpolate.uniform(
            shape_native=(2, 2),
            pixel_scales=2.0,
            pixel_scales_interp=0.1,
            store_slim=False,
        )

        assert type(grid) == grids.Grid2DInterpolate
        assert type(grid.slim) == grids.Grid2DInterpolate
        assert type(grid.native) == grids.Grid2DInterpolate
        assert (
            grid == np.array([[[1.0, -1.0], [1.0, 1.0]], [[-1.0, -1.0], [-1.0, 1.0]]])
        ).all()
        assert (
            grid.native
            == np.array([[[1.0, -1.0], [1.0, 1.0]], [[-1.0, -1.0], [-1.0, 1.0]]])
        ).all()
        assert (
            grid.slim == np.array([[1.0, -1.0], [1.0, 1.0], [-1.0, -1.0], [-1.0, 1.0]])
        ).all()
        assert grid.pixel_scales == (2.0, 2.0)
        assert grid.pixel_scales_interp == (0.1, 0.1)
        assert grid.origin == (0.0, 0.0)

        grid = aa.Grid2DInterpolate.uniform(
            shape_native=(2, 1), pixel_scales=1.0, pixel_scales_interp=0.2, sub_size=2
        )

        assert type(grid) == grids.Grid2DInterpolate
        assert type(grid.slim) == grids.Grid2DInterpolate
        assert type(grid.native) == grids.Grid2DInterpolate
        assert (
            grid.native
            == np.array(
                [
                    [[0.75, -0.25], [0.75, 0.25]],
                    [[0.25, -0.25], [0.25, 0.25]],
                    [[-0.25, -0.25], [-0.25, 0.25]],
                    [[-0.75, -0.25], [-0.75, 0.25]],
                ]
            )
        ).all()
        assert (
            grid.slim
            == np.array(
                [
                    [0.75, -0.25],
                    [0.75, 0.25],
                    [0.25, -0.25],
                    [0.25, 0.25],
                    [-0.25, -0.25],
                    [-0.25, 0.25],
                    [-0.75, -0.25],
                    [-0.75, 0.25],
                ]
            )
        ).all()
        assert (grid.native_binned == np.array([[[0.5, 0.0]], [[-0.5, 0.0]]])).all()
        assert (grid.slim_binned == np.array([[0.5, 0.0], [-0.5, 0.0]])).all()
        assert grid.pixel_scales == (1.0, 1.0)
        assert grid.pixel_scales_interp == (0.2, 0.2)
        assert grid.origin == (0.0, 0.0)
        assert grid.sub_size == 2



import autoarray as aa
import numpy as np
import scipy.spatial


class TestRectangular:
    class TestNeighbors:
        def test__3x3_grid(self):

            # I0I1I2I
            # I3I4I5I
            # I6I7I8I

            (
                pixel_neighbors,
                pixel_neighbors_size,
            ) = aa.util.pixelization.rectangular_neighbors_from(shape_native=(3, 3))

            assert (pixel_neighbors[0] == [1, 3, -1, -1]).all()
            assert (pixel_neighbors[1] == [0, 2, 4, -1]).all()
            assert (pixel_neighbors[2] == [1, 5, -1, -1]).all()
            assert (pixel_neighbors[3] == [0, 4, 6, -1]).all()
            assert (pixel_neighbors[4] == [1, 3, 5, 7]).all()
            assert (pixel_neighbors[5] == [2, 4, 8, -1]).all()
            assert (pixel_neighbors[6] == [3, 7, -1, -1]).all()
            assert (pixel_neighbors[7] == [4, 6, 8, -1]).all()
            assert (pixel_neighbors[8] == [5, 7, -1, -1]).all()

            assert (pixel_neighbors_size == np.array([2, 3, 2, 3, 4, 3, 2, 3, 2])).all()

        def test__3x4_grid(self):

            # I0I1I 2I 3I
            # I4I5I 6I 7I
            # I8I9I10I11I

            (
                pixel_neighbors,
                pixel_neighbors_size,
            ) = aa.util.pixelization.rectangular_neighbors_from(shape_native=(3, 4))

            assert (pixel_neighbors[0] == [1, 4, -1, -1]).all()
            assert (pixel_neighbors[1] == [0, 2, 5, -1]).all()
            assert (pixel_neighbors[2] == [1, 3, 6, -1]).all()
            assert (pixel_neighbors[3] == [2, 7, -1, -1]).all()
            assert (pixel_neighbors[4] == [0, 5, 8, -1]).all()
            assert (pixel_neighbors[5] == [1, 4, 6, 9]).all()
            assert (pixel_neighbors[6] == [2, 5, 7, 10]).all()
            assert (pixel_neighbors[7] == [3, 6, 11, -1]).all()
            assert (pixel_neighbors[8] == [4, 9, -1, -1]).all()
            assert (pixel_neighbors[9] == [5, 8, 10, -1]).all()
            assert (pixel_neighbors[10] == [6, 9, 11, -1]).all()
            assert (pixel_neighbors[11] == [7, 10, -1, -1]).all()

            assert (
                pixel_neighbors_size == np.array([2, 3, 3, 2, 3, 4, 4, 3, 2, 3, 3, 2])
            ).all()

        def test__4x3_grid(self):

            # I0I 1I 2I
            # I3I 4I 5I
            # I6I 7I 8I
            # I9I10I11I

            (
                pixel_neighbors,
                pixel_neighbors_size,
            ) = aa.util.pixelization.rectangular_neighbors_from(shape_native=(4, 3))

            assert (pixel_neighbors[0] == [1, 3, -1, -1]).all()
            assert (pixel_neighbors[1] == [0, 2, 4, -1]).all()
            assert (pixel_neighbors[2] == [1, 5, -1, -1]).all()
            assert (pixel_neighbors[3] == [0, 4, 6, -1]).all()
            assert (pixel_neighbors[4] == [1, 3, 5, 7]).all()
            assert (pixel_neighbors[5] == [2, 4, 8, -1]).all()
            assert (pixel_neighbors[6] == [3, 7, 9, -1]).all()
            assert (pixel_neighbors[7] == [4, 6, 8, 10]).all()
            assert (pixel_neighbors[8] == [5, 7, 11, -1]).all()
            assert (pixel_neighbors[9] == [6, 10, -1, -1]).all()
            assert (pixel_neighbors[10] == [7, 9, 11, -1]).all()
            assert (pixel_neighbors[11] == [8, 10, -1, -1]).all()

            assert (
                pixel_neighbors_size == np.array([2, 3, 2, 3, 4, 3, 3, 4, 3, 2, 3, 2])
            ).all()

        def test__4x4_grid(self):

            # I0 I 1I 2I 3I
            # I4 I 5I 6I 7I
            # I8 I 9I10I11I
            # I12I13I14I15I

            (
                pixel_neighbors,
                pixel_neighbors_size,
            ) = aa.util.pixelization.rectangular_neighbors_from(shape_native=(4, 4))

            assert (pixel_neighbors[0] == [1, 4, -1, -1]).all()
            assert (pixel_neighbors[1] == [0, 2, 5, -1]).all()
            assert (pixel_neighbors[2] == [1, 3, 6, -1]).all()
            assert (pixel_neighbors[3] == [2, 7, -1, -1]).all()
            assert (pixel_neighbors[4] == [0, 5, 8, -1]).all()
            assert (pixel_neighbors[5] == [1, 4, 6, 9]).all()
            assert (pixel_neighbors[6] == [2, 5, 7, 10]).all()
            assert (pixel_neighbors[7] == [3, 6, 11, -1]).all()
            assert (pixel_neighbors[8] == [4, 9, 12, -1]).all()
            assert (pixel_neighbors[9] == [5, 8, 10, 13]).all()
            assert (pixel_neighbors[10] == [6, 9, 11, 14]).all()
            assert (pixel_neighbors[11] == [7, 10, 15, -1]).all()
            assert (pixel_neighbors[12] == [8, 13, -1, -1]).all()
            assert (pixel_neighbors[13] == [9, 12, 14, -1]).all()
            assert (pixel_neighbors[14] == [10, 13, 15, -1]).all()
            assert (pixel_neighbors[15] == [11, 14, -1, -1]).all()

            assert (
                pixel_neighbors_size
                == np.array([2, 3, 3, 2, 3, 4, 4, 3, 3, 4, 4, 3, 2, 3, 3, 2])
            ).all()


class TestVoronoi:
    class TestNeighbors:
        def test__points_in_x_cross_shape__neighbors_of_each_pixel_correct(self):

            points = np.array(
                [[1.0, -1.0], [1.0, 1.0], [0.0, 0.0], [-1.0, -1.0], [-1.0, 1.0]]
            )

            voronoi = scipy.spatial.Voronoi(points, qhull_options="Qbb Qc Qx Qm")
            (
                pixel_neighbors,
                pixel_neighbors_size,
            ) = aa.util.pixelization.voronoi_neighbors_from(
                pixels=5, ridge_points=np.array(voronoi.ridge_points)
            )

            assert set(pixel_neighbors[0]) == {1, 2, 3, -1}
            assert set(pixel_neighbors[1]) == {0, 2, 4, -1}
            assert set(pixel_neighbors[2]) == {0, 1, 3, 4}
            assert set(pixel_neighbors[3]) == {0, 2, 4, -1}
            assert set(pixel_neighbors[4]) == {1, 2, 3, -1}

            assert (pixel_neighbors_size == np.array([3, 3, 4, 3, 3])).all()

        def test__9_points_in_square___neighbors_of_each_pixel_correct(self):

            # 9 points in a square - makes a square (this is the example int he scipy documentaiton page)

            points = np.array(
                [
                    [2.0, 0.0],
                    [2.0, 1.0],
                    [2.0, 2.0],
                    [1.0, 0.0],
                    [1.0, 1.0],
                    [1.0, 2.0],
                    [0.0, 0.0],
                    [0.0, 1.0],
                    [0.0, 2.0],
                ]
            )

            voronoi = scipy.spatial.Voronoi(points, qhull_options="Qbb Qc Qx Qm")
            (
                pixel_neighbors,
                pixel_neighbors_size,
            ) = aa.util.pixelization.voronoi_neighbors_from(
                pixels=9, ridge_points=np.array(voronoi.ridge_points)
            )

            assert set(pixel_neighbors[0]) == {1, 3, -1, -1}
            assert set(pixel_neighbors[1]) == {0, 2, 4, -1}
            assert set(pixel_neighbors[2]) == {1, 5, -1, -1}
            assert set(pixel_neighbors[3]) == {0, 4, 6, -1}
            assert set(pixel_neighbors[4]) == {1, 3, 5, 7}
            assert set(pixel_neighbors[5]) == {2, 4, 8, -1}
            assert set(pixel_neighbors[6]) == {3, 7, -1, -1}
            assert set(pixel_neighbors[7]) == {4, 6, 8, -1}
            assert set(pixel_neighbors[8]) == {5, 7, -1, -1}

            assert (pixel_neighbors_size == np.array([2, 3, 2, 3, 4, 3, 2, 3, 2])).all()



import numpy as np

from autoarray.structures.arrays import abstract_array
from autoarray.structures import frames as f
from autoarray.structures import region as reg
from autoarray.util import frame_util


class AbstractFrame2D(abstract_array.AbstractArray2D):
    def __array_finalize__(self, obj):

        super(AbstractFrame2D, self).__array_finalize__(obj)

        if isinstance(obj, AbstractFrame2D):
            if hasattr(obj, "roe_corner"):
                self.roe_corner = obj.roe_corner

            if hasattr(obj, "original_roe_corner"):
                self.original_roe_corner = obj.original_roe_corner

            if hasattr(obj, "scans"):
                self.scans = obj.scans

            if hasattr(obj, "exposure_info"):
                self.exposure_info = obj.exposure_info

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(AbstractFrame2D, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    # noinspection PyMethodOverriding
    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super(AbstractFrame2D, self).__setstate__(state[0:-1])

    def _new_structure(self, array, mask, store_slim):
        return self.__class__(
            array=array,
            mask=mask,
            original_roe_corner=self.original_roe_corner,
            scans=self.scans,
            exposure_info=self.exposure_info,
        )

    @property
    def original_orientation(self):
        return frame_util.rotate_array_from_roe_corner(
            array=self, roe_corner=self.original_roe_corner
        )

    @property
    def binned_across_parallel(self):
        return np.mean(np.ma.masked_array(self, self.mask), axis=0)

    @property
    def binned_across_serial(self):
        return np.mean(np.ma.masked_array(self, self.mask), axis=1)

    @property
    def parallel_overscan_frame(self):
        """Extract an arrays of all of the parallel trails in the parallel overscan region, that are to the side of a
        charge-injection scans from a charge injection ci_frame.

        The diagram below illustrates the arrays that is extracted from a ci_frame:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = parallel prescan       [ssssssssss] = parallel overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / parallel charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
          [...][ccccccccccccccccccccc][sts]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][000000000000000000000][000]
          [000][000000000000000000000][tst]
        | [000][000000000000000000000][sts]    |
        | [000][000000000000000000000][000]    | Direction
        P [000][000000000000000000000][000]    | of
        | [000][000000000000000000000][tst]    | clocking
          [000][000000000000000000000][sts]    |

        []     [=====================]
               <---------S----------
        """
        return f.Frame2D.extracted_frame_from_frame_and_extraction_region(
            frame=self, extraction_region=self.scans.parallel_overscan
        )

    @property
    def parallel_overscan_binned_line(self):
        return self.parallel_overscan_frame.binned_across_serial

    def parallel_trail_from_y(self, y, dy):
        """Grid2DIrregularGrouped of a parallel trail of size dy from coordinate y"""
        return (int(y - dy), int(y + 1))

    def serial_trail_from_x(self, x, dx):
        """Grid2DIrregularGrouped of a serial trail of size dx from coordinate x"""
        return (int(x), int(x + 1 + dx))

    def parallel_front_edge_of_region(self, region, rows):

        reg.check_parallel_front_edge_size(region=region, rows=rows)

        y_coord = region.y0
        y_min = y_coord + rows[0]
        y_max = y_coord + rows[1]

        return reg.Region2D((y_min, y_max, region.x0, region.x1))

    def parallel_trails_of_region(self, region, rows=(0, 1)):
        y_coord = region.y1
        y_min = y_coord + rows[0]
        y_max = y_coord + rows[1]
        return reg.Region2D((y_min, y_max, region.x0, region.x1))

    def parallel_side_nearest_read_out_region(self, region, columns=(0, 1)):
        x_min, x_max = self.x_limits(region, columns)
        return reg.Region2D(region=(0, self.shape_native[0], x_min, x_max))

    @property
    def serial_overscan_frame(self):
        """Extract an arrays of all of the serial trails in the serial overscan region, that are to the side of a
        charge-injection scans from a charge injection ci_frame.

        The diagram below illustrates the arrays that is extracted from a ci_frame:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
          [...][ccccccccccccccccccccc][sts]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][000000000000000000000][000]
          [000][000000000000000000000][tst]
        | [000][000000000000000000000][sts]    |
        | [000][000000000000000000000][000]    | Direction
        P [000][000000000000000000000][000]    | of
        | [000][000000000000000000000][tst]    | clocking
          [000][000000000000000000000][sts]    |

        []     [=====================]
               <---------S----------
        """
        return f.Frame2D.extracted_frame_from_frame_and_extraction_region(
            frame=self, extraction_region=self.scans.serial_overscan
        )

    @property
    def serial_overscan_binned_line(self):
        return self.serial_overscan_frame.binned_across_parallel

    def serial_front_edge_of_region(self, region, columns=(0, 1)):
        reg.check_serial_front_edge_size(region, columns)
        x_min, x_max = self.x_limits(region, columns)
        return reg.Region2D(region=(region.y0, region.y1, x_min, x_max))

    def serial_trails_of_region(self, region, columns=(0, 1)):
        x_coord = region.x1
        x_min = x_coord + columns[0]
        x_max = x_coord + columns[1]
        return reg.Region2D(region=(region.y0, region.y1, x_min, x_max))

    def serial_entire_rows_of_region(self, region):
        return reg.Region2D(region=(region.y0, region.y1, 0, self.shape_native[1]))

    def x_limits(self, region, columns):
        x_coord = region.x0
        x_min = x_coord + columns[0]
        x_max = x_coord + columns[1]
        return x_min, x_max


class Scans:
    def __init__(
        self, parallel_overscan=None, serial_prescan=None, serial_overscan=None
    ):

        if isinstance(parallel_overscan, tuple):
            parallel_overscan = reg.Region2D(region=parallel_overscan)

        if isinstance(serial_prescan, tuple):
            serial_prescan = reg.Region2D(region=serial_prescan)

        if isinstance(serial_overscan, tuple):
            serial_overscan = reg.Region2D(region=serial_overscan)

        self.parallel_overscan = parallel_overscan
        self.serial_prescan = serial_prescan
        self.serial_overscan = serial_overscan

    @classmethod
    def rotated_from_roe_corner(cls, roe_corner, shape_native, scans):

        if scans is None:
            return Scans()

        parallel_overscan = frame_util.rotate_region_from_roe_corner(
            region=scans.parallel_overscan,
            shape_native=shape_native,
            roe_corner=roe_corner,
        )
        serial_prescan = frame_util.rotate_region_from_roe_corner(
            region=scans.serial_prescan,
            shape_native=shape_native,
            roe_corner=roe_corner,
        )
        serial_overscan = frame_util.rotate_region_from_roe_corner(
            region=scans.serial_overscan,
            shape_native=shape_native,
            roe_corner=roe_corner,
        )

        return Scans(
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
            serial_overscan=serial_overscan,
        )

    @classmethod
    def after_extraction(cls, frame, extraction_region):

        parallel_overscan = frame_util.region_after_extraction(
            original_region=frame.scans.parallel_overscan,
            extraction_region=extraction_region,
        )
        serial_prescan = frame_util.region_after_extraction(
            original_region=frame.scans.serial_prescan,
            extraction_region=extraction_region,
        )
        serial_overscan = frame_util.region_after_extraction(
            original_region=frame.scans.serial_overscan,
            extraction_region=extraction_region,
        )

        return Scans(
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
            serial_overscan=serial_overscan,
        )

    @classmethod
    def from_frame(cls, frame):

        return Scans(
            parallel_overscan=frame.scans.parallel_overscan,
            serial_prescan=frame.scans.serial_prescan,
            serial_overscan=frame.scans.serial_overscan,
        )

    @property
    def serial_trails_columns(self):
        return self.serial_overscan[3] - self.serial_overscan[2]



from autoarray.plot.plotters import abstract_plotters
from autoarray.plot.mat_wrap import visuals as vis
from autoarray.plot.mat_wrap import include as inc
from autoarray.plot.mat_wrap import mat_plot as mp
from autoarray.structures import arrays, frames, grids, lines
from autoarray.inversion import mappers
import typing


class Array2DPlotter(abstract_plotters.AbstractPlotter):
    def __init__(
        self,
        array: arrays.Array2D,
        mat_plot_2d: mp.MatPlot2D = mp.MatPlot2D(),
        visuals_2d: vis.Visuals2D = vis.Visuals2D(),
        include_2d: inc.Include2D = inc.Include2D(),
    ):

        super().__init__(
            visuals_2d=visuals_2d, include_2d=include_2d, mat_plot_2d=mat_plot_2d
        )

        self.array = array

    @property
    def visuals_with_include_2d(self):
        """
        Extracts from an `Array2D` attributes that can be plotted and returns them in a `Visuals` object.

        Only attributes already in `self.visuals_2d` or with `True` entries in the `Include` object are extracted
        for plotting.

        From an `Array2D` the following attributes can be extracted for plotting:

        - origin: the (y,x) origin of the structure's coordinate system.
        - mask: the mask of the structure.
        - border: the border of the structure's mask.

        Parameters
        ----------
        array : arrays.Array2D
            The array whose attributes are extracted for plotting.

        Returns
        -------
        vis.Visuals2D
            The collection of attributes that can be plotted by a `Plotter2D` object.
        """
        return self.visuals_2d + self.visuals_2d.__class__(
            origin=self.extract_2d(
                "origin", grids.Grid2DIrregular(grid=[self.array.origin])
            ),
            mask=self.extract_2d("mask", self.array.mask),
            border=self.extract_2d(
                "border", self.array.mask.border_grid_sub_1.slim_binned
            ),
        )

    def figure(self):

        self.mat_plot_2d.plot_array(
            array=self.array,
            visuals_2d=self.visuals_with_include_2d,
            auto_labels=mp.AutoLabels(title="Array2D", filename="array"),
        )


class Frame2DPlotter(abstract_plotters.AbstractPlotter):
    def __init__(
        self,
        frame: frames.Frame2D,
        mat_plot_2d: mp.MatPlot2D = mp.MatPlot2D(),
        visuals_2d: vis.Visuals2D = vis.Visuals2D(),
        include_2d: inc.Include2D = inc.Include2D(),
    ):
        super().__init__(
            visuals_2d=visuals_2d, include_2d=include_2d, mat_plot_2d=mat_plot_2d
        )

        self.frame = frame

    @property
    def visuals_with_include_2d(self):
        """
        Extracts from a `Frame2D` attributes that can be plotted and return them in a `Visuals` object.

        Only attributes already in `self.visuals_2d` or with `True` entries in the `Include` object are extracted
        for plotting.

        From an `Frame2D` the following attributes can be extracted for plotting:

        - origin: the (y,x) origin of the structure's coordinate system.
        - mask: the mask of the structure.
        - border: the border of the structure's mask.
        - parallel_overscan: the parallel overscan of the frame data.
        - serial_prescan: the serial prescan of the frame data.
        - serial_overscan: the serial overscan of the frame data.

        Parameters
        ----------
        frame : frames.Frame2D
            The frame whose attributes are extracted for plotting.

        Returns
        -------
        vis.Visuals2D
            The collection of attributes that can be plotted by a `Plotter2D` object.
        """
        return self.visuals_2d + self.visuals_2d.__class__(
            origin=self.extract_2d(
                "origin", grids.Grid2DIrregular(grid=[self.frame.origin])
            ),
            mask=self.extract_2d("mask", self.frame.mask),
            border=self.extract_2d(
                "border", self.frame.mask.border_grid_sub_1.slim_binned
            ),
            parallel_overscan=self.extract_2d(
                "parallel_overscan", self.frame.scans.parallel_overscan
            ),
            serial_prescan=self.extract_2d(
                "serial_prescan", self.frame.scans.serial_prescan
            ),
            serial_overscan=self.extract_2d(
                "serial_overscan", self.frame.scans.serial_overscan
            ),
        )

    def figure(self):

        self.mat_plot_2d.plot_frame(
            frame=self.frame,
            visuals_2d=self.visuals_with_include_2d,
            auto_labels=mp.AutoLabels(title="Frame2D", filename="frame"),
        )


class Grid2DPlotter(abstract_plotters.AbstractPlotter):
    def __init__(
        self,
        grid: grids.Grid2D,
        mat_plot_2d: mp.MatPlot2D = mp.MatPlot2D(),
        visuals_2d: vis.Visuals2D = vis.Visuals2D(),
        include_2d: inc.Include2D = inc.Include2D(),
    ):
        super().__init__(
            visuals_2d=visuals_2d, include_2d=include_2d, mat_plot_2d=mat_plot_2d
        )

        self.grid = grid

    @property
    def visuals_with_include_2d(self):
        """
        Extracts from a `Grid2D` attributes that can be plotted and return them in a `Visuals` object.

        Only attributes with `True` entries in the `Include` object are extracted for plotting.

        From a `Grid2D` the following attributes can be extracted for plotting:

        - origin: the (y,x) origin of the grid's coordinate system.
        - mask: the mask of the grid.
        - border: the border of the grid's mask.

        Parameters
        ----------
        grid : abstract_grid.AbstractGrid2D
            The grid whose attributes are extracted for plotting.

        Returns
        -------
        vis.Visuals2D
            The collection of attributes that can be plotted by a `Plotter2D` object.
        """
        if not isinstance(self.grid, grids.Grid2D):
            return self.visuals_2d

        return self.visuals_2d + self.visuals_2d.__class__(
            origin=self.extract_2d(
                "origin", grids.Grid2DIrregular(grid=[self.grid.origin])
            )
        )

    def figure(self, color_array=None):

        self.mat_plot_2d.plot_grid(
            grid=self.grid,
            visuals_2d=self.visuals_with_include_2d,
            auto_labels=mp.AutoLabels(title="Grid2D", filename="grid"),
            color_array=color_array,
        )


class MapperPlotter(abstract_plotters.AbstractPlotter):
    def __init__(
        self,
        mapper: typing.Union[mappers.MapperRectangular, mappers.MapperVoronoi],
        mat_plot_2d: mp.MatPlot2D = mp.MatPlot2D(),
        visuals_2d: vis.Visuals2D = vis.Visuals2D(),
        include_2d: inc.Include2D = inc.Include2D(),
    ):
        super().__init__(
            visuals_2d=visuals_2d, include_2d=include_2d, mat_plot_2d=mat_plot_2d
        )

        self.mapper = mapper

    @property
    def visuals_data_with_include_2d(self):
        """
        Extracts from a `Mapper` attributes that can be plotted for figures in its data-plane (e.g. the reconstructed
        data) and return them in a `Visuals` object.

        Only attributes with `True` entries in the `Include` object are extracted for plotting.

        From a `Mapper` the following attributes can be extracted for plotting in the data-plane:

        - origin: the (y,x) origin of the `Array2D`'s coordinate system in the data plane.
        - mask : the `Mask` defined in the data-plane containing the data that is used by the `Mapper`.
        - mapper_data_pixelization_grid: the `Mapper`'s pixelization grid in the data-plane.
        - mapper_border_grid: the border of the `Mapper`'s full grid in the data-plane.

        Parameters
        ----------
        mapper : mappers.Mapper
            The mapper whose data-plane attributes are extracted for plotting.

        Returns
        -------
        vis.Visuals2D
            The collection of attributes that can be plotted by a `Plotter2D` object.
        """
        return self.visuals_2d + self.visuals_2d.__class__(
            origin=self.extract_2d(
                "origin",
                grids.Grid2DIrregular(grid=[self.mapper.source_grid_slim.mask.origin]),
            ),
            mask=self.extract_2d("mask", self.mapper.source_grid_slim.mask),
            border=self.extract_2d(
                "border",
                self.mapper.source_grid_slim.mask.border_grid_sub_1.slim_binned,
            ),
            pixelization_grid=self.extract_2d(
                "pixelization_grid",
                self.mapper.data_pixelization_grid,
                "mapper_data_pixelization_grid",
            ),
        )

    @property
    def visuals_source_with_include_2d(self):
        """
        Extracts from a `Mapper` attributes that can be plotted for figures in its source-plane (e.g. the reconstruction
        and return them in a `Visuals` object.

        Only attributes with `True` entries in the `Include` object are extracted for plotting.

        From a `Mapper` the following attributes can be extracted for plotting in the source-plane:

        - origin: the (y,x) origin of the coordinate system in the source plane.
        - mapper_source_pixelization_grid: the `Mapper`'s pixelization grid in the source-plane.
        - mapper_source_grid_slim: the `Mapper`'s full grid in the source-plane.
        - mapper_border_grid: the border of the `Mapper`'s full grid in the data-plane.

        Parameters
        ----------
        mapper : mappers.Mapper
            The mapper whose source-plane attributes are extracted for plotting.

        Returns
        -------
        vis.Visuals2D
            The collection of attributes that can be plotted by a `Plotter2D` object.
        """

        return self.visuals_2d + self.visuals_2d.__class__(
            origin=self.extract_2d(
                "origin",
                grids.Grid2DIrregular(
                    grid=[self.mapper.source_pixelization_grid.origin]
                ),
            ),
            grid=self.extract_2d(
                "grid", self.mapper.source_grid_slim, "mapper_source_grid_slim"
            ),
            border=self.extract_2d(
                "border", self.mapper.source_grid_slim.sub_border_grid
            ),
            pixelization_grid=self.extract_2d(
                "pixelization_grid",
                self.mapper.source_pixelization_grid,
                "mapper_source_pixelization_grid",
            ),
        )

    def figure(self, source_pixelilzation_values=None):

        self.mat_plot_2d.plot_mapper(
            mapper=self.mapper,
            visuals_2d=self.visuals_source_with_include_2d,
            source_pixelilzation_values=source_pixelilzation_values,
            auto_labels=mp.AutoLabels(title="Mapper", filename="mapper"),
        )

    def subplot_image_and_mapper(self, image):

        self.open_subplot_figure(number_subplots=2)

        self.mat_plot_2d.plot_array(
            array=image,
            visuals_2d=self.visuals_data_with_include_2d,
            auto_labels=mp.AutoLabels(title="Image"),
        )

        if self.visuals_2d.pixelization_indexes is not None:

            indexes = self.mapper.slim_indexes_from_pixelization_indexes(
                pixelization_indexes=self.visuals_2d.pixelization_indexes
            )

            self.mat_plot_2d.index_scatter.scatter_grid_indexes(
                grid=self.mapper.source_grid_slim.mask.masked_grid, indexes=indexes
            )

        self.figure()

        self.mat_plot_2d.output.subplot_to_figure(
            auto_filename="subplot_image_and_mapper"
        )
        self.close_subplot_figure()


class Line1DPlotter(abstract_plotters.AbstractPlotter):
    def __init__(
        self,
        y,
        x,
        mat_plot_1d: mp.MatPlot1D = mp.MatPlot1D(),
        visuals_1d: vis.Visuals1D = vis.Visuals1D(),
        include_1d: inc.Include1D = inc.Include1D(),
    ):

        super().__init__(
            visuals_1d=visuals_1d, include_1d=include_1d, mat_plot_1d=mat_plot_1d
        )

    def visuals_from_line(self, line: lines.Line1D) -> "vis.Visuals1D":

        origin = line.origin if self.include_1d.origin else None
        mask = line.mask if self.include_1d.mask else None

        return vis.Visuals1D(origin=origin, mask=mask)

    def figure(
        self,
        y,
        x,
        label=None,
        plot_axis_type="semilogy",
        vertical_lines=None,
        vertical_line_labels=None,
    ):

        self.mat_plot_1d.plot_line(
            y=y,
            x=x,
            visuals_1d=self.visuals_1d,
            auto_labels=mp.AutoLabels(),
            label=label,
            plot_axis_type=plot_axis_type,
            vertical_lines=vertical_lines,
            vertical_line_labels=vertical_line_labels,
        )



import numpy as np
import scipy.spatial
import scipy.spatial.qhull as qhull

from autoarray import exc
from autoarray.structures import grids
from autoarray.mask import mask_2d as msk
from autoarray.util import grid_util, pixelization_util


class Grid2DRectangular(grids.Grid2D):
    def __new__(
        cls, grid, shape_native, pixel_scales, origin=(0.0, 0.0), *args, **kwargs
    ):
        """A grid of (y,x) coordinates which reprsent a rectangular grid of pixels which are used to form the pixel centres of adaptive pixelizations in the \
        *pixelizations* module.

        A `Grid2DRectangular` is ordered such pixels begin from the top-row of the mask and go rightwards and then \
        downwards. Therefore, it is a ndarray of shape [total_pix_pixels, 2]. The first element of the ndarray \
        thus corresponds to the pixelization pixel index and second element the y or x arc -econd coordinates. For example:

        - pix_grid[3,0] = the 4th unmasked pixel's y-coordinate.
        - pix_grid[6,1] = the 7th unmasked pixel's x-coordinate.

        Parameters
        -----------
        pix_grid : np.ndarray
            The grid of (y,x) scaled coordinates of every image-plane pixelization grid used for adaptive source \
            -plane pixelizations.
        nearest_pixelization_index_for_slim_index : np.ndarray
            A 1D array that maps every grid pixel to its nearest pixelization-grid pixel.
        """

        mask = msk.Mask2D.unmasked(
            shape_native=shape_native,
            pixel_scales=pixel_scales,
            sub_size=1,
            origin=origin,
        )

        obj = super().__new__(cls=cls, grid=grid, mask=mask)

        (
            pixel_neighbors,
            pixel_neighbors_size,
        ) = pixelization_util.rectangular_neighbors_from(shape_native=shape_native)
        obj.pixel_neighbors = pixel_neighbors.astype("int")
        obj.pixel_neighbors_size = pixel_neighbors_size.astype("int")
        return obj

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super().__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    # noinspection PyMethodOverriding
    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super().__setstate__(state[0:-1])

    @classmethod
    def overlay_grid(cls, shape_native, grid, buffer=1e-8):
        """The geometry of a rectangular grid.

        This is used to map grid of (y,x) scaled coordinates to the pixels on the rectangular grid.

        Parameters
        -----------
        shape_native : (int, int)
            The dimensions of the rectangular grid of pixels (y_pixels, x_pixel)
        pixel_scales : (float, float)
            The pixel conversion scale of a pixel in the y and x directions.
        origin : (float, float)
            The scaled origin of the rectangular pixelization's coordinate system.
        pixel_neighbors : np.ndarray
            An array of length (y_pixels*x_pixels) which provides the index of all neighbors of every pixel in \
            the rectangular grid (entries of -1 correspond to no neighbor).
        pixel_neighbors_size : ndarrayy
            An array of length (y_pixels*x_pixels) which gives the number of neighbors of every pixel in the \
            rectangular grid.
        """

        y_min = np.min(grid[:, 0]) - buffer
        y_max = np.max(grid[:, 0]) + buffer
        x_min = np.min(grid[:, 1]) - buffer
        x_max = np.max(grid[:, 1]) + buffer

        pixel_scales = (
            float((y_max - y_min) / shape_native[0]),
            float((x_max - x_min) / shape_native[1]),
        )

        origin = ((y_max + y_min) / 2.0, (x_max + x_min) / 2.0)

        grid_slim = grid_util.grid_2d_slim_via_shape_native_from(
            shape_native=shape_native,
            pixel_scales=pixel_scales,
            sub_size=1,
            origin=origin,
        )

        return Grid2DRectangular(
            grid=grid_slim,
            shape_native=shape_native,
            pixel_scales=pixel_scales,
            origin=origin,
        )

    @property
    def pixels(self):
        return self.shape_native[0] * self.shape_native[1]

    @property
    def shape_native_scaled(self):
        return (
            (self.shape_native[0] * self.pixel_scales[0]),
            (self.shape_native[1] * self.pixel_scales[1]),
        )


class Grid2DVoronoi(np.ndarray):
    """
    Returns the geometry of the Voronoi pixelization, by alligning it with the outer-most coordinates on a \
    grid plus a small buffer.

    Parameters
    -----------
    grid : np.ndarray
        The (y,x) grid of coordinates which determine the Voronoi pixelization's
    pixelization_grid : np.ndarray
        The (y,x) centre of every Voronoi pixel in scaleds.
    origin : (float, float)
        The scaled origin of the Voronoi pixelization's coordinate system.
    pixel_neighbors : np.ndarray
        An array of length (voronoi_pixels) which provides the index of all neighbors of every pixel in \
        the Voronoi grid (entries of -1 correspond to no neighbor).
    pixel_neighbors_size : ndarrayy
        An array of length (voronoi_pixels) which gives the number of neighbors of every pixel in the \
        Voronoi grid.
    """

    def __new__(
        cls, grid, nearest_pixelization_index_for_slim_index=None, *args, **kwargs
    ):
        """A pixelization-grid of (y,x) coordinates which are used to form the pixel centres of adaptive pixelizations in the \
        *pixelizations* module.

        A `Grid2DRectangular` is ordered such pixels begin from the top-row of the mask and go rightwards and then \
        downwards. Therefore, it is a ndarray of shape [total_pix_pixels, 2]. The first element of the ndarray \
        thus corresponds to the pixelization pixel index and second element the y or x arc -econd coordinates. For example:

        - pix_grid[3,0] = the 4th unmasked pixel's y-coordinate.
        - pix_grid[6,1] = the 7th unmasked pixel's x-coordinate.

        Parameters
        -----------
        pix_grid : np.ndarray
            The grid of (y,x) scaled coordinates of every image-plane pixelization grid used for adaptive source \
            -plane pixelizations.
        nearest_pixelization_index_for_slim_index : np.ndarray
            A 1D array that maps every grid pixel to its nearest pixelization-grid pixel.
        """

        if type(grid) is list:
            grid = np.asarray(grid)

        obj = grid.view(cls)
        obj.nearest_pixelization_index_for_slim_index = (
            nearest_pixelization_index_for_slim_index
        )

        try:
            obj.voronoi = scipy.spatial.Voronoi(
                np.asarray([grid[:, 1], grid[:, 0]]).T, qhull_options="Qbb Qc Qx Qm"
            )
        except (ValueError, OverflowError, scipy.spatial.qhull.QhullError) as e:
            raise exc.PixelizationException() from e

        (
            pixel_neighbors,
            pixel_neighbors_size,
        ) = pixelization_util.voronoi_neighbors_from(
            pixels=obj.pixels, ridge_points=np.asarray(obj.voronoi.ridge_points)
        )

        obj.pixel_neighbors = pixel_neighbors.astype("int")
        obj.pixel_neighbors_size = pixel_neighbors_size.astype("int")
        obj.nearest_pixelization_index_for_slim_index = (
            nearest_pixelization_index_for_slim_index
        )

        return obj

    def __array_finalize__(self, obj):

        if hasattr(obj, "voronoi"):
            self.voronoi = obj.voronoi

        if hasattr(obj, "pixel_neighbors"):
            self.pixel_neighbors = obj.pixel_neighbors

        if hasattr(obj, "pixel_neighbors_size"):
            self.pixel_neighbors_size = obj.pixel_neighbors_size

        if hasattr(obj, "nearest_pixelization_index_for_slim_index"):
            self.nearest_pixelization_index_for_slim_index = (
                obj.nearest_pixelization_index_for_slim_index
            )

        if hasattr(obj, "_sub_border_flat_indexes"):
            self._sub_border_flat_indexes = obj._sub_border_flat_indexes

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super().__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    # noinspection PyMethodOverriding
    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super().__setstate__(state[0:-1])

    @property
    def origin(self):
        return 0.0, 0.0

    @property
    def pixels(self):
        return self.shape[0]

    @property
    def sub_border_grid(self):
        """The (y,x) grid of all sub-pixels which are at the border of the mask.

        This is NOT all sub-pixels which are in mask pixels at the mask's border, but specifically the sub-pixels
        within these border pixels which are at the extreme edge of the border."""
        return self[self._sub_border_flat_indexes]

    @classmethod
    def manual_slim(cls, grid):
        return Grid2DVoronoi(grid=grid)

    @classmethod
    def from_grid_and_unmasked_2d_grid_shape(cls, unmasked_sparse_shape, grid):

        sparse_grid = grids.Grid2DSparse.from_grid_and_unmasked_2d_grid_shape(
            unmasked_sparse_shape=unmasked_sparse_shape, grid=grid
        )

        return Grid2DVoronoi(
            grid=sparse_grid.sparse,
            nearest_pixelization_index_for_slim_index=sparse_grid.sparse_index_for_slim_index,
        )

    @property
    def shape_native_scaled(self):
        return (
            np.amax(self[:, 0]) - np.amin(self[:, 0]),
            np.amax(self[:, 1]) - np.amin(self[:, 1]),
        )

    @property
    def scaled_maxima(self):
        return (np.amax(self[:, 0]), np.amax(self[:, 1]))

    @property
    def scaled_minima(self):
        return (np.amin(self[:, 0]), np.amin(self[:, 1]))

    @property
    def extent(self):
        return np.asarray(
            [
                self.scaled_minima[1],
                self.scaled_maxima[1],
                self.scaled_minima[0],
                self.scaled_maxima[0],
            ]
        )



import autoarray as aa
import numpy as np


class TestBinnedPaddingArray:
    def test__bin_up_factor_is_1__array_2d_does_not_change_shape(self):

        array_2d = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=1
        )

        assert (array_2d_padded == array_2d).all()

    def test__bin_up_factor_gives_no_remainder__array_2d_does_not_change_shape(self):

        array_2d = np.ones(shape=(6, 6))

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=2
        )
        assert array_2d_padded.shape == (6, 6)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=3
        )
        assert array_2d_padded.shape == (6, 6)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=6
        )
        assert array_2d_padded.shape == (6, 6)

        array_2d = np.ones(shape=(8, 8))

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=2
        )
        assert array_2d_padded.shape == (8, 8)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=4
        )
        assert array_2d_padded.shape == (8, 8)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=8
        )
        assert array_2d_padded.shape == (8, 8)

        array_2d = np.ones(shape=(9, 9))

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=3
        )
        assert array_2d_padded.shape == (9, 9)

        array_2d = np.ones(shape=(16, 16))

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=2
        )
        assert array_2d_padded.shape == (16, 16)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=4
        )
        assert array_2d_padded.shape == (16, 16)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=8
        )
        assert array_2d_padded.shape == (16, 16)

        array_2d = np.ones(shape=(12, 16))

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=2
        )
        assert array_2d_padded.shape == (12, 16)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=4
        )
        assert array_2d_padded.shape == (12, 16)

        array_2d = np.ones(shape=(16, 12))

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=2
        )
        assert array_2d_padded.shape == (16, 12)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=4
        )
        assert array_2d_padded.shape == (16, 12)

    def test__bin_up_factor_gives_remainder__array_2d_padded_to_give_no_remainder(self):

        array_2d = np.ones(shape=(6, 6))

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=4
        )
        assert array_2d_padded.shape == (8, 8)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=5
        )
        assert array_2d_padded.shape == (10, 10)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=7
        )
        assert array_2d_padded.shape == (7, 7)

        array_2d = np.ones(shape=(10, 10))

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=3
        )
        assert array_2d_padded.shape == (12, 12)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=4
        )
        assert array_2d_padded.shape == (12, 12)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=6
        )
        assert array_2d_padded.shape == (12, 12)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=7
        )
        assert array_2d_padded.shape == (14, 14)

        array_2d = np.ones(shape=(7, 10))

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=3
        )
        assert array_2d_padded.shape == (9, 12)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=5
        )
        assert array_2d_padded.shape == (10, 10)

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=7
        )
        assert array_2d_padded.shape == (7, 14)

    def test__padding_using_arrays_and_not_shapes(self):

        array_2d = np.ones(shape=(4, 4))
        array_2d[1, 1] = 2.0

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=3
        )
        assert array_2d_padded.shape == (6, 6)
        assert (
            array_2d_padded
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                    [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
                    [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                    [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=5
        )
        assert array_2d_padded.shape == (5, 5)
        assert (
            array_2d_padded
            == np.array(
                [
                    [1.0, 1.0, 1.0, 1.0, 0.0],
                    [1.0, 2.0, 1.0, 1.0, 0.0],
                    [1.0, 1.0, 1.0, 1.0, 0.0],
                    [1.0, 1.0, 1.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        array_2d = np.ones(shape=(2, 3))
        array_2d[1, 1] = 2.0
        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=2
        )
        assert array_2d_padded.shape == (2, 4)
        assert (
            array_2d_padded == np.array([[0.0, 1.0, 1.0, 1.0], [0.0, 1.0, 2.0, 1.0]])
        ).all()

        array_2d = np.ones(shape=(3, 2))
        array_2d[1, 1] = 2.0
        array_2d_padded = aa.util.binning.padded_binning_array_2d_from(
            array_2d=array_2d, bin_up_factor=2
        )
        assert array_2d_padded.shape == (4, 2)
        assert (
            array_2d_padded
            == np.array([[0.0, 0.0], [1.0, 1.0], [1.0, 2.0], [1.0, 1.0]])
        ).all()


class TestBinnedArrays2d:
    def test__bin_using_mean__array_4x4_to_2x2__uses_mean_correctly(self):

        array_2d = np.array(
            [
                [1.0, 1.0, 2.0, 2.0],
                [1.0, 1.0, 2.0, 2.0],
                [3.0, 3.0, 4.0, 4.0],
                [3.0, 3.0, 4.0, 4.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_mean(
            array_2d=array_2d, bin_up_factor=2
        )

        assert (binned_array_2d == np.array([[1.0, 2.0], [3.0, 4.0]])).all()

        array_2d = np.array(
            [
                [1.0, 2.0, 2.0, 2.0],
                [1.0, 6.0, 2.0, 10.0],
                [9.0, 3.0, 4.0, 0.0],
                [3.0, 3.0, 4.0, 4.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_mean(
            array_2d=array_2d, bin_up_factor=2
        )

        assert (binned_array_2d == np.array([[2.5, 4.0], [4.5, 3.0]])).all()

    def test__bin_using_mean__array_6x3_to_2x1_and_3x6_to_1x2__uses_mean_correctly(
        self,
    ):

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_mean(
            array_2d=array_2d, bin_up_factor=3
        )

        assert (binned_array_2d == np.array([[1.0], [2.0]])).all()

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0],
                [1.0, 10.0, 1.0],
                [1.0, 1.0, 1.0],
                [2.0, 11.0, 2.0],
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_mean(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (binned_array_2d == np.array([[2.0], [3.0]])).all()

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_mean(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (binned_array_2d == np.array([[1.0, 2.0]])).all()

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                [1.0, 10.0, 1.0, 11.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_mean(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (binned_array_2d == np.array([[2.0, 3.0]])).all()

    def test__bin_using_mean__bin_includes_padding_image_with_zeros(self):

        # Padded array:

        # [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

        array_2d = np.ones(shape=(4, 4))
        array_2d[1, 1] = 2.0
        binned_array_2d = aa.util.binning.bin_array_2d_via_mean(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (
            binned_array_2d
            == np.array([[(5.0 / 9.0), (4.0 / 9.0)], [(4.0 / 9.0), (4.0 / 9.0)]])
        ).all()

        # Padded Array2D:

        # np.array([[0.0, 1.0, 1.0, 1.0],
        #           [0.0, 1.0, 2.0, 1.0]]

        array_2d = np.ones(shape=(2, 3))
        array_2d[1, 1] = 2.0
        binned_2d_array = aa.util.binning.bin_array_2d_via_mean(
            array_2d=array_2d, bin_up_factor=2
        )
        assert (binned_2d_array == np.array([[0.5, 1.25]])).all()

    def test__bin_using_quadrature__array_4x4_to_2x2__uses_quadrature_correctly(self):

        array_2d = np.array(
            [
                [1.0, 1.0, 2.0, 2.0],
                [1.0, 1.0, 2.0, 2.0],
                [3.0, 3.0, 4.0, 4.0],
                [3.0, 3.0, 4.0, 4.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_quadrature(
            array_2d=array_2d, bin_up_factor=2
        )

        assert (
            binned_array_2d
            == np.array(
                [
                    [np.sqrt(4.0) / 4.0, np.sqrt(16.0) / 4.0],
                    [np.sqrt(36.0) / 4.0, np.sqrt(64.0) / 4.0],
                ]
            )
        ).all()

        array_2d = np.array(
            [
                [1.0, 2.0, 2.0, 2.0],
                [1.0, 6.0, 2.0, 10.0],
                [9.0, 3.0, 4.0, 0.0],
                [3.0, 3.0, 4.0, 4.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_quadrature(
            array_2d=array_2d, bin_up_factor=2
        )

        assert (
            binned_array_2d
            == np.array(
                [
                    [np.sqrt(42.0) / 4.0, np.sqrt(112.0) / 4.0],
                    [np.sqrt(108.0) / 4.0, np.sqrt(48.0) / 4.0],
                ]
            )
        ).all()

    def test__bin_using_quadrature__array_6x3_to_2x1_and_3x6_to_1x2__uses_quadrature_correctly(
        self,
    ):

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_quadrature(
            array_2d=array_2d, bin_up_factor=3
        )

        assert (
            binned_array_2d == np.array([[np.sqrt(9.0) / 9.0], [np.sqrt(36.0) / 9.0]])
        ).all()

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0],
                [1.0, 10.0, 1.0],
                [1.0, 1.0, 1.0],
                [2.0, 4.0, 2.0],
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_quadrature(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (
            binned_array_2d == np.array([[np.sqrt(108.0) / 9.0], [np.sqrt(48.0) / 9.0]])
        ).all()

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_quadrature(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (
            binned_array_2d == np.array([[np.sqrt(9.0) / 9.0, np.sqrt(36.0) / 9.0]])
        ).all()

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                [1.0, 10.0, 1.0, 4.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_quadrature(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (
            binned_array_2d == np.array([[np.sqrt(108.0) / 9.0, np.sqrt(48.0) / 9.0]])
        ).all()

    def test__bin_using_quadrature__bin_includes_padding_image_with_zeros(self):

        # Padded array:

        # [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

        array_2d = np.ones(shape=(4, 4))
        array_2d[1, 1] = 2.0
        binned_array_2d = aa.util.binning.bin_array_2d_via_quadrature(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (
            binned_array_2d
            == np.array(
                [
                    [np.sqrt(7.0) / 9.0, np.sqrt(4.0) / 9.0],
                    [np.sqrt(4.0) / 9.0, np.sqrt(4.0) / 9.0],
                ]
            )
        ).all()

        # Padded Array2D:

        # np.array([[0.0, 1.0, 1.0, 1.0],
        #           [0.0, 1.0, 2.0, 1.0]]

        array_2d = np.ones(shape=(2, 3))
        array_2d[1, 1] = 2.0
        binned_2d_array = aa.util.binning.bin_array_2d_via_quadrature(
            array_2d=array_2d, bin_up_factor=2
        )
        assert (
            binned_2d_array == np.array([[np.sqrt(2.0) / 4.0, np.sqrt(7.0) / 4.0]])
        ).all()

    def test__bin_using_sum__array_4x4_to_2x2__uses_sum_correctly(self):

        array_2d = np.array(
            [
                [1.0, 1.0, 2.0, 2.0],
                [1.0, 1.0, 2.0, 2.0],
                [3.0, 3.0, 4.0, 4.0],
                [3.0, 3.0, 4.0, 4.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_sum(
            array_2d=array_2d, bin_up_factor=2
        )

        assert (binned_array_2d == np.array([[4.0, 8.0], [12.0, 16.0]])).all()

        array_2d = np.array(
            [
                [1.0, 2.0, 2.0, 2.0],
                [1.0, 6.0, 2.0, 10.0],
                [9.0, 3.0, 4.0, 0.0],
                [3.0, 3.0, 4.0, 4.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_sum(
            array_2d=array_2d, bin_up_factor=2
        )

        assert (binned_array_2d == np.array([[10.0, 16.0], [18.0, 12.0]])).all()

    def test__bin_using_sum__array_6x3_to_2x1_and_3x6_to_1x2__uses_sum_correctly(self):

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_sum(
            array_2d=array_2d, bin_up_factor=3
        )

        assert (binned_array_2d == np.array([[9.0], [18.0]])).all()

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0],
                [1.0, 10.0, 1.0],
                [1.0, 1.0, 1.0],
                [2.0, 11.0, 2.0],
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_sum(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (binned_array_2d == np.array([[18.0], [27.0]])).all()

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_sum(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (binned_array_2d == np.array([[9.0, 18.0]])).all()

        array_2d = np.array(
            [
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
                [1.0, 10.0, 1.0, 11.0, 2.0, 2.0],
                [1.0, 1.0, 1.0, 2.0, 2.0, 2.0],
            ]
        )

        binned_array_2d = aa.util.binning.bin_array_2d_via_sum(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (binned_array_2d == np.array([[18.0, 27.0]])).all()

    def test__bin_using_sum__bin_includes_padding_image_with_zeros(self):

        # Padded array:

        # [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

        array_2d = np.ones(shape=(4, 4))
        array_2d[1, 1] = 2.0
        binned_array_2d = aa.util.binning.bin_array_2d_via_sum(
            array_2d=array_2d, bin_up_factor=3
        )
        assert (binned_array_2d == np.array([[5.0, 4.0], [4.0, 4.0]])).all()

        # Padded Array2D:

        # np.array([[0.0, 1.0, 1.0, 1.0],
        #           [0.0, 1.0, 2.0, 1.0]]

        array_2d = np.ones(shape=(2, 3))
        array_2d[1, 1] = 2.0
        binned_2d_array = aa.util.binning.bin_array_2d_via_sum(
            array_2d=array_2d, bin_up_factor=2
        )
        assert (binned_2d_array == np.array([[2.0, 5.0]])).all()


class TestBinUpMask2d:
    def test__mask_4x4_to_2x2__creates_correct_binned_up_mask(self):

        mask = np.array(
            [
                [True, False, True, True],
                [True, True, True, True],
                [True, True, False, False],
                [False, True, True, True],
            ]
        )

        binned_mask = aa.util.binning.bin_mask(mask=mask, bin_up_factor=2)

        assert (binned_mask == np.array([[False, True], [False, False]])).all()

        mask = np.array(
            [
                [True, True, True, True],
                [True, True, True, True],
                [True, True, False, False],
                [True, True, True, True],
            ]
        )

        binned_mask = aa.util.binning.bin_mask(mask=mask, bin_up_factor=2)

        assert (binned_mask == np.array([[True, True], [True, False]])).all()

    def test__mask_6x3_to_2x1_and_3x6_to_1x2__sets_up_correct_mask(self):

        mask = np.array(
            [
                [True, True, True],
                [True, True, True],
                [True, True, True],
                [True, True, True],
                [True, True, True],
                [True, True, True],
            ]
        )

        binned_mask = aa.util.binning.bin_mask(mask=mask, bin_up_factor=3)

        assert (binned_mask == np.array([[True], [True]])).all()

        mask = np.array(
            [
                [True, True, True],
                [True, True, False],
                [True, True, True],
                [True, True, True],
                [True, True, True],
                [True, True, True],
            ]
        )

        binned_mask = aa.util.binning.bin_mask(mask=mask, bin_up_factor=3)
        assert (binned_mask == np.array([[False], [True]])).all()

        mask = np.array(
            [
                [True, True, True, True, True, True],
                [True, True, True, True, True, True],
                [True, True, True, True, True, True],
            ]
        )

        binned_mask = aa.util.binning.bin_mask(mask=mask, bin_up_factor=3)
        assert (binned_mask == np.array([[True, True]])).all()

        mask = np.array(
            [
                [True, True, True, True, True, True],
                [True, True, True, True, True, True],
                [True, True, True, True, True, False],
            ]
        )

        binned_mask = aa.util.binning.bin_mask(mask=mask, bin_up_factor=3)
        assert (binned_mask == np.array([[True, False]])).all()

    def test__bin_includes_padding_image_with_zeros(self):
        # Padded mask:

        # [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 2.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
        #  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

        mask = np.full(shape=(4, 4), fill_value=True)
        mask[1, 1] = False
        mask[3, 3] = False
        binned_mask = aa.util.binning.bin_mask(mask=mask, bin_up_factor=3)
        assert (binned_mask == np.array([[False, True], [True, False]])).all()

        # Padded Array2D:

        # np.array([[0.0, 1.0, 1.0, 1.0],
        #           [0.0, 1.0, 2.0, 1.0]]

        mask = np.full(shape=(2, 3), fill_value=True)
        mask[1, 2] = False
        binned_2d_mask = aa.util.binning.bin_mask(mask=mask, bin_up_factor=2)
        assert (binned_2d_mask == np.array([[True, False]])).all()


class TestMask1DIndexForPaddedMask2dIndex:
    def test__no_padding__mask_is_full_of_false__returns_indexes_in_ascending_order(
        self,
    ):

        mask = np.full(fill_value=False, shape=(4, 4))

        mask_1d_index_for_padded_mask_index = aa.util.binning.mask_1d_index_for_padded_mask_index_from(
            mask=mask, bin_up_factor=1
        )

        assert (
            mask_1d_index_for_padded_mask_index
            == np.array([[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11], [12, 13, 14, 15]])
        ).all()

        mask = np.array([[False, False], [True, False], [True, False]])

        mask_1d_index_for_padded_mask_index = aa.util.binning.mask_1d_index_for_padded_mask_index_from(
            mask=mask, bin_up_factor=1
        )

        assert (
            mask_1d_index_for_padded_mask_index == np.array([[0, 1], [-1, 2], [-1, 3]])
        ).all()

    def test__includes_padding__padded_entries_are_given_minus_ones(self):

        mask = np.full(fill_value=False, shape=(4, 4))

        mask_1d_index_for_padded_mask_index = aa.util.binning.mask_1d_index_for_padded_mask_index_from(
            mask=mask, bin_up_factor=3
        )

        assert (
            mask_1d_index_for_padded_mask_index
            == np.array(
                [
                    [-1, -1, -1, -1, -1, -1],
                    [-1, 0, 1, 2, 3, -1],
                    [-1, 4, 5, 6, 7, -1],
                    [-1, 8, 9, 10, 11, -1],
                    [-1, 12, 13, 14, 15, -1],
                    [-1, -1, -1, -1, -1, -1],
                ]
            )
        ).all()

        mask = np.array([[False, False], [True, False], [True, False]])

        mask_1d_index_for_padded_mask_index = aa.util.binning.mask_1d_index_for_padded_mask_index_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            mask_1d_index_for_padded_mask_index
            == np.array([[-1, -1], [0, 1], [-1, 2], [-1, 3]])
        ).all()


class TestBinnedMask1dIndexesForMask2dIndex:
    def test__masks_are_full_arrays_and_bin_up_factor_2__mapping_is_correct(self):

        mask = np.full(fill_value=False, shape=(4, 4))

        binned_mask_1d_index_for_mask_index = aa.util.binning.binned_mask_1d_index_for_padded_mask_index_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            binned_mask_1d_index_for_mask_index
            == np.array([[0, 0, 1, 1], [0, 0, 1, 1], [2, 2, 3, 3], [2, 2, 3, 3]])
        ).all()

        mask = np.full(fill_value=False, shape=(9, 9))

        binned_mask_1d_index_for_mask_index = aa.util.binning.binned_mask_1d_index_for_padded_mask_index_from(
            mask=mask, bin_up_factor=3
        )

        assert (
            binned_mask_1d_index_for_mask_index
            == np.array(
                [
                    [0, 0, 0, 1, 1, 1, 2, 2, 2],
                    [0, 0, 0, 1, 1, 1, 2, 2, 2],
                    [0, 0, 0, 1, 1, 1, 2, 2, 2],
                    [3, 3, 3, 4, 4, 4, 5, 5, 5],
                    [3, 3, 3, 4, 4, 4, 5, 5, 5],
                    [3, 3, 3, 4, 4, 4, 5, 5, 5],
                    [6, 6, 6, 7, 7, 7, 8, 8, 8],
                    [6, 6, 6, 7, 7, 7, 8, 8, 8],
                    [6, 6, 6, 7, 7, 7, 8, 8, 8],
                ]
            )
        ).all()

    def test__masks_are_rectangular_arrays__include_areas_which_bin_up_is_all_true(
        self,
    ):

        mask = np.array(
            [
                [True, False, True, True, True, True],
                [False, False, False, True, True, True],
            ]
        )

        binned_mask_1d_index_for_mask_index = aa.util.binning.binned_mask_1d_index_for_padded_mask_index_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            binned_mask_1d_index_for_mask_index
            == np.array([[-1, 0, -1, -1, -1, -1], [0, 0, 1, -1, -1, -1]])
        ).all()

        mask = np.array(
            [
                [True, False],
                [False, False],
                [False, True],
                [True, True],
                [True, True],
                [True, True],
            ]
        )

        binned_mask_1d_index_for_mask_index = aa.util.binning.binned_mask_1d_index_for_padded_mask_index_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            binned_mask_1d_index_for_mask_index
            == np.array([[-1, 0], [0, 0], [1, -1], [-1, -1], [-1, -1], [-1, -1]])
        ).all()

    def test__mask_includes_padding__mapper_mask_accounts_for_padding(self):

        mask = np.full(fill_value=False, shape=(5, 5))

        binned_mask_1d_index_for_mask_index = aa.util.binning.binned_mask_1d_index_for_padded_mask_index_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            binned_mask_1d_index_for_mask_index
            == np.array(
                [
                    [-1, -1, -1, -1, -1, -1],
                    [-1, 0, 1, 1, 2, 2],
                    [-1, 3, 4, 4, 5, 5],
                    [-1, 3, 4, 4, 5, 5],
                    [-1, 6, 7, 7, 8, 8],
                    [-1, 6, 7, 7, 8, 8],
                ]
            )
        ).all()


class TestArray1DToBininedArray1d:
    def test__masks_are_full_arrays_and_bin_up_factor_2__mapping_is_correct(self):

        mask = np.full(fill_value=False, shape=(4, 4))

        binned_masked_array_1d_for_masked_array_1d = aa.util.binning.binned_masked_array_1d_for_masked_array_1d_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            binned_masked_array_1d_for_masked_array_1d
            == np.array([0, 0, 1, 1, 0, 0, 1, 1, 2, 2, 3, 3, 2, 2, 3, 3])
        ).all()

        mask = np.full(fill_value=False, shape=(9, 9))

        binned_masked_array_1d_for_masked_array_1d = aa.util.binning.binned_masked_array_1d_for_masked_array_1d_from(
            mask=mask, bin_up_factor=3
        )

        assert (
            binned_masked_array_1d_for_masked_array_1d
            == np.array(
                [
                    0,
                    0,
                    0,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    0,
                    0,
                    0,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    0,
                    0,
                    0,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                    6,
                    6,
                    6,
                    7,
                    7,
                    7,
                    8,
                    8,
                    8,
                    6,
                    6,
                    6,
                    7,
                    7,
                    7,
                    8,
                    8,
                    8,
                    6,
                    6,
                    6,
                    7,
                    7,
                    7,
                    8,
                    8,
                    8,
                ]
            )
        ).all()

    def test__masks_are_rectangular_arrays__include_areas_which_bin_up_is_all_true(
        self,
    ):

        mask = np.array(
            [
                [True, False, True, True, True, True],
                [False, False, False, True, True, True],
            ]
        )

        binned_masked_array_1d_for_masked_array_1d = aa.util.binning.binned_masked_array_1d_for_masked_array_1d_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            binned_masked_array_1d_for_masked_array_1d == np.array([0, 0, 0, 1])
        ).all()

        mask = np.array(
            [
                [True, False],
                [False, False],
                [False, True],
                [True, True],
                [True, False],
                [True, False],
            ]
        )

        binned_masked_array_1d_for_masked_array_1d = aa.util.binning.binned_masked_array_1d_for_masked_array_1d_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            binned_masked_array_1d_for_masked_array_1d == np.array([0, 0, 0, 1, 2, 2])
        ).all()

    def test__mask_includes_padding__mapper_mask_accounts_for_padding(self):

        mask = np.full(fill_value=False, shape=(5, 5))

        binned_masked_array_1d_for_masked_array_1d = aa.util.binning.binned_masked_array_1d_for_masked_array_1d_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            binned_masked_array_1d_for_masked_array_1d
            == np.array(
                [
                    0,
                    1,
                    1,
                    2,
                    2,
                    3,
                    4,
                    4,
                    5,
                    5,
                    3,
                    4,
                    4,
                    5,
                    5,
                    6,
                    7,
                    7,
                    8,
                    8,
                    6,
                    7,
                    7,
                    8,
                    8,
                ]
            )
        ).all()


class TestBinnedMaskArrayToArray:
    def test__masks_are_full_arrays_and_bin_up_factor_2__mapping_is_correct(self):

        mask = np.full(fill_value=False, shape=(4, 4))

        masked_array_1d_for_binned_masked_array_1d = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            masked_array_1d_for_binned_masked_array_1d == np.array([0, 2, 8, 10])
        ).all()

        mask = np.full(fill_value=False, shape=(9, 9))

        masked_array_1d_for_binned_masked_array_1d = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_from(
            mask=mask, bin_up_factor=3
        )

        assert (
            masked_array_1d_for_binned_masked_array_1d
            == np.array([0, 3, 6, 27, 30, 33, 54, 57, 60])
        ).all()

    def test__masks_are_rectangular_arrays__include_areas_which_bin_up_is_all_true(
        self,
    ):

        mask = np.array(
            [
                [True, False, True, True, True, True],
                [False, False, False, True, True, True],
            ]
        )

        masked_array_1d_for_binned_masked_array_1d = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_from(
            mask=mask, bin_up_factor=2
        )

        assert (masked_array_1d_for_binned_masked_array_1d == np.array([0, 3])).all()

        mask = np.array(
            [
                [True, False],
                [False, False],
                [False, True],
                [True, True],
                [True, True],
                [True, True],
            ]
        )

        masked_array_1d_for_binned_masked_array_1d = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_from(
            mask=mask, bin_up_factor=2
        )

        assert (masked_array_1d_for_binned_masked_array_1d == np.array([0, 3])).all()

    def test__mask_includes_padding__mapper_mask_accounts_for_padding(self):

        mask = np.full(fill_value=False, shape=(5, 5))

        masked_array_1d_for_binned_masked_array_1d = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            masked_array_1d_for_binned_masked_array_1d
            == np.array([0, 1, 3, 5, 6, 8, 15, 16, 18])
        ).all()


class TestBinnedMaskArrayToArrayAll:
    def test__masks_are_full_arrays_and_bin_up_factor_2__mapping_is_correct(self):

        mask = np.full(fill_value=False, shape=(4, 4))

        (
            masked_array_1d_for_binned_masked_array_1d_all,
            binned_masked_array_1d_sizes,
        ) = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_all_from(
            mask=mask, bin_up_factor=2
        )

        assert (binned_masked_array_1d_sizes == np.array([4, 4, 4, 4])).all()

        assert (
            masked_array_1d_for_binned_masked_array_1d_all
            == np.array([[0, 1, 4, 5], [2, 3, 6, 7], [8, 9, 12, 13], [10, 11, 14, 15]])
        ).all()

        mask = np.full(fill_value=False, shape=(9, 9))

        (
            masked_array_1d_for_binned_masked_array_1d_all,
            binned_masked_array_1d_sizes,
        ) = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_all_from(
            mask=mask, bin_up_factor=3
        )

        assert (
            binned_masked_array_1d_sizes == np.array([9, 9, 9, 9, 9, 9, 9, 9, 9])
        ).all()

        assert (
            masked_array_1d_for_binned_masked_array_1d_all
            == np.array(
                [
                    [0, 1, 2, 9, 10, 11, 18, 19, 20],
                    [3, 4, 5, 12, 13, 14, 21, 22, 23],
                    [6, 7, 8, 15, 16, 17, 24, 25, 26],
                    [27, 28, 29, 36, 37, 38, 45, 46, 47],
                    [30, 31, 32, 39, 40, 41, 48, 49, 50],
                    [33, 34, 35, 42, 43, 44, 51, 52, 53],
                    [54, 55, 56, 63, 64, 65, 72, 73, 74],
                    [57, 58, 59, 66, 67, 68, 75, 76, 77],
                    [60, 61, 62, 69, 70, 71, 78, 79, 80],
                ]
            )
        ).all()

    def test__masks_are_rectangular_arrays__include_areas_which_bin_up_is_all_true(
        self,
    ):

        mask = np.array(
            [
                [True, False, True, True, True, True],
                [False, False, False, True, True, True],
            ]
        )

        (
            masked_array_1d_for_binned_masked_array_1d_all,
            binned_masked_array_1d_sizes,
        ) = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_all_from(
            mask=mask, bin_up_factor=2
        )

        assert (binned_masked_array_1d_sizes == np.array([3, 1])).all()

        assert (
            masked_array_1d_for_binned_masked_array_1d_all
            == np.array([[0, 1, 2, -1], [3, -1, -1, -1]])
        ).all()

        mask = np.array(
            [
                [True, False],
                [False, False],
                [False, True],
                [True, True],
                [True, True],
                [True, True],
            ]
        )

        (
            masked_array_1d_for_binned_masked_array_1d_all,
            binned_masked_array_1d_sizes,
        ) = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_all_from(
            mask=mask, bin_up_factor=2
        )

        assert (binned_masked_array_1d_sizes == np.array([3, 1])).all()

        assert (
            masked_array_1d_for_binned_masked_array_1d_all
            == np.array([[0, 1, 2, -1], [3, -1, -1, -1]])
        ).all()

    def test__mask_includes_padding__mapper_mask_accounts_for_padding(self):

        mask = np.full(fill_value=False, shape=(5, 5))

        (
            masked_array_1d_for_binned_masked_array_1d_all,
            binned_masked_array_1d_sizes,
        ) = aa.util.binning.masked_array_1d_for_binned_masked_array_1d_all_from(
            mask=mask, bin_up_factor=2
        )

        assert (
            binned_masked_array_1d_sizes == np.array([1, 2, 2, 2, 4, 4, 2, 4, 4])
        ).all()

        assert (
            masked_array_1d_for_binned_masked_array_1d_all
            == np.array(
                [
                    [0, -1, -1, -1],
                    [1, 2, -1, -1],
                    [3, 4, -1, -1],
                    [5, 10, -1, -1],
                    [6, 7, 11, 12],
                    [8, 9, 13, 14],
                    [15, 20, -1, -1],
                    [16, 17, 21, 22],
                    [18, 19, 23, 24],
                ]
            )
        ).all()
